from typing import List, NamedTuple
from xml.etree.ElementTree import Element

from loxun import XmlWriter

from .utils import enforce_not_null


class Group(NamedTuple):
    id: int
    name: str
    children: List["Group"]  # type: ignore[misc]

    def _dump(self, xf: XmlWriter) -> None:
        xf.startTag("group", {"id": str(self.id), "name": self.name})
        for subgroup in self.children:
            subgroup._dump(xf)
        xf.endTag()  # group

    @classmethod
    def _parse(cls, nml_group: Element) -> "Group":
        return cls(
            int(enforce_not_null(nml_group.get("id"))),
            enforce_not_null(nml_group.get("name")),
            [],
        )



from enum import Enum, unique
from typing import Union


@unique
class SamplingModes(Enum):
    ANISOTROPIC = "anisotropic"
    ISOTROPIC = "isotropic"
    CONSTANT_Z = "constant_z"

    @classmethod
    def parse(cls, mode: Union[str, "SamplingModes"]) -> "SamplingModes":
        if mode == "auto":
            return cls.ANISOTROPIC
        else:
            return cls(mode)



"""This module converts a RAW dataset to a WEBKNOSSOS dataset."""

import logging
from argparse import Namespace
from functools import partial
from multiprocessing import cpu_count
from pathlib import Path
from typing import Any, Literal, Optional, Tuple, Union

import numpy as np
import typer
from typing_extensions import Annotated

from ..dataset import DataFormat, Dataset, MagView, SamplingModes
from ..dataset.defaults import DEFAULT_CHUNK_SHAPE, DEFAULT_CHUNKS_PER_SHARD
from ..geometry import BoundingBox, Mag, Vec3Int
from ..utils import (
    get_executor_for_args,
    time_start,
    time_stop,
    wait_and_ensure_success,
)
from ._utils import (
    DistributionStrategy,
    Order,
    SamplingMode,
    VoxelSize,
    parse_mag,
    parse_path,
    parse_vec3int,
    parse_voxel_size,
)

logger = logging.getLogger(__name__)


def _raw_chunk_converter(
    bounding_box: BoundingBox,
    source_raw_path: Path,
    target_mag_view: MagView,
    input_dtype: str,
    shape: Tuple[int, int, int],
    order: Literal["C", "F"],
    flip_axes: Optional[Union[int, Tuple[int, ...]]],
) -> None:
    logging.info("Conversion of %s", bounding_box.topleft)
    source_data: np.ndarray = np.memmap(
        source_raw_path,
        dtype=np.dtype(input_dtype),
        mode="r",
        shape=(1,) + shape,
        order=order,
    )

    if flip_axes:
        source_data = np.flip(source_data, flip_axes)

    contiguous_chunk = source_data[(slice(None),) + bounding_box.to_slices()].copy(
        order="F"
    )
    target_mag_view.write(contiguous_chunk, bounding_box.topleft)


def convert_raw(
    source_raw_path: Path,
    target_path: Path,
    layer_name: str,
    input_dtype: str,
    shape: Vec3Int,
    data_format: DataFormat,
    chunk_shape: Vec3Int,
    chunks_per_shard: Vec3Int,
    order: Literal["C", "F"] = "F",
    voxel_size: Optional[Tuple[float, float, float]] = (1.0, 1.0, 1.0),
    flip_axes: Optional[Union[int, Tuple[int, ...]]] = None,
    compress: bool = True,
    executor_args: Optional[Namespace] = None,
) -> MagView:
    """Performs the conversion step from RAW file to WEBKNOSSOS"""
    time_start(f"Conversion of {source_raw_path}")

    if voxel_size is None:
        voxel_size = 1.0, 1.0, 1.0
    wk_ds = Dataset(target_path, voxel_size=voxel_size, exist_ok=True)
    wk_layer = wk_ds.get_or_add_layer(
        layer_name,
        "color",
        dtype_per_layer=np.dtype(input_dtype),
        num_channels=1,
        data_format=data_format,
    )
    wk_layer.bounding_box = BoundingBox((0, 0, 0), shape)
    wk_mag = wk_layer.get_or_add_mag(
        "1",
        chunk_shape=chunk_shape,
        chunks_per_shard=chunks_per_shard,
        compress=compress,
    )

    # Parallel chunk conversion
    with get_executor_for_args(args=executor_args) as executor:
        wait_and_ensure_success(
            executor.map_to_futures(
                partial(
                    _raw_chunk_converter,
                    source_raw_path=source_raw_path,
                    target_mag_view=wk_mag,
                    input_dtype=input_dtype,
                    shape=shape,
                    order=order,
                    flip_axes=flip_axes,
                ),
                wk_layer.bounding_box.chunk(chunk_shape=chunk_shape * chunks_per_shard),
            ),
            executor=executor,
        )

    time_stop(f"Conversion of {source_raw_path}")
    return wk_mag


def main(
    *,
    source: Annotated[
        Any,
        typer.Argument(
            help="Path to your raw image data.",
            show_default=False,
            parser=parse_path,
        ),
    ],
    target: Annotated[
        Any,
        typer.Argument(
            help="Target path to save your WEBKNOSSOS dataset.",
            show_default=False,
            parser=parse_path,
        ),
    ],
    shape: Annotated[
        Vec3Int,
        typer.Option(
            help="Shape of the source dataset. Should be a comma seperated "
            "sting (e.g. 1024,1024,512).",
            parser=parse_vec3int,
            metavar="Vec3Int",
        ),
    ],
    order: Annotated[
        Order,
        typer.Option(
            help="The input data storage layout: "
            "either 'F' for Fortran-style/column-major order (the default), "
            "or 'C' for C-style/row-major order. "
            "Note: Axes are expected in  (x, y, z) order."
        ),
    ] = Order.F,
    layer_name: Annotated[
        str,
        typer.Option(help="Name of the cubed layer (color or segmentation)"),
    ] = "color",
    voxel_size: Annotated[
        Optional[VoxelSize],
        typer.Option(
            help="The size of one voxel in source data in nanometers. "
            "Should be a comma seperated string (e.g. 11.0,11.0,20.0).",
            parser=parse_voxel_size,
            metavar="VOXEL_SIZE",
        ),
    ] = None,
    dtype: Annotated[
        str, typer.Option(help="Target datatype (e.g. uint8, uint16, uint32)")
    ] = "uint8",
    data_format: Annotated[
        DataFormat,
        typer.Option(
            help="Data format to store the target dataset in.",
        ),
    ] = "wkw",  # type:ignore
    chunk_shape: Annotated[
        Vec3Int,
        typer.Option(
            help="Number of voxels to be stored as a chunk in the output format "
            "(e.g. `32` or `32,32,32`).",
            parser=parse_vec3int,
            metavar="Vec3Int",
        ),
    ] = DEFAULT_CHUNK_SHAPE,
    chunks_per_shard: Annotated[
        Vec3Int,
        typer.Option(
            help="Number of chunks to be stored as a shard in the output format "
            "(e.g. `32` or `32,32,32`).",
            parser=parse_vec3int,
            metavar="Vec3Int",
        ),
    ] = DEFAULT_CHUNKS_PER_SHARD,
    max_mag: Annotated[
        Optional[Mag],
        typer.Option(
            help="Max resolution to be downsampled. "
            "Should be number or minus seperated string (e.g. 2 or 2-2-2).",
            parser=parse_mag,
        ),
    ] = None,
    interpolation_mode: Annotated[
        str,
        typer.Option(
            help="The interpolation mode that should be used "
            "(median, mode, nearest, bilinear or bicubic)."
        ),
    ] = "default",
    flip_axes: Annotated[
        Optional[Vec3Int],
        typer.Option(
            help="The axes at which should be flipped. "
            "Input format is a comma separated list of axis indices. "
            "For example, 1,2,3 will flip the x, y and z axes.",
            parser=parse_vec3int,
            metavar="Vec3Int",
        ),
    ] = None,
    compress: Annotated[
        bool, typer.Option(help="Enable compression of the target dataset.")
    ] = False,
    sampling_mode: Annotated[
        SamplingMode, typer.Option(help="The sampling mode to use.")
    ] = SamplingMode.ANISOTROPIC,
    jobs: Annotated[
        int,
        typer.Option(
            help="Number of processes to be spawned.",
            rich_help_panel="Executor options",
        ),
    ] = cpu_count(),
    distribution_strategy: Annotated[
        DistributionStrategy,
        typer.Option(
            help="Strategy to distribute the task across CPUs or nodes.",
            rich_help_panel="Executor options",
        ),
    ] = DistributionStrategy.MULTIPROCESSING,
    job_resources: Annotated[
        Optional[str],
        typer.Option(
            help='Necessary when using slurm as distribution strategy. Should be a JSON string \
(e.g., --job_resources=\'{"mem": "10M"}\')\'',
            rich_help_panel="Executor options",
        ),
    ] = None,
) -> None:
    """Converts a RAW file into a WEBKNOSSOS dataset."""

    if flip_axes is not None:
        for index in flip_axes:
            assert (
                0 <= index <= 3
            ), "flip_axes parameter must only contain indices between 0 and 3."

    mode = SamplingModes.parse(sampling_mode.value)
    if source.is_dir():
        logger.error("source_path is not a file")
        return

    executor_args = Namespace(
        jobs=jobs,
        distribution_strategy=distribution_strategy.value,
        job_resources=job_resources,
    )

    mag_view = convert_raw(
        source,
        target,
        layer_name,
        dtype,
        shape,
        data_format,
        chunk_shape,
        chunks_per_shard,
        order.value,
        voxel_size,
        flip_axes,
        compress,
        executor_args,
    )

    with get_executor_for_args(executor_args) as executor:
        mag_view.layer.downsample(
            from_mag=mag_view.mag,
            coarsest_mag=max_mag,
            interpolation_mode=interpolation_mode,
            compress=compress,
            sampling_mode=mode,
            executor=executor,
        )



import os

from ..geometry import Vec3Int
from .data_format import DataFormat

DEFAULT_WKW_FILE_LEN = 32
DEFAULT_CHUNK_SHAPE = Vec3Int.full(32)
DEFAULT_DATA_FORMAT = (
    DataFormat(os.environ["WK_DEFAULT_DATA_FORMAT"])
    if "WK_DEFAULT_DATA_FORMAT" in os.environ
    else DataFormat.WKW
)
DEFAULT_CHUNKS_PER_SHARD = (
    Vec3Int.from_str(os.environ["WK_DEFAULT_CHUNKS_PER_SHARD"])
    if "WK_DEFAULT_CHUNKS_PER_SHARD" in os.environ
    else Vec3Int.full(32)
)
DEFAULT_CHUNKS_PER_SHARD_ZARR = Vec3Int.full(1)
DEFAULT_CHUNKS_PER_SHARD_FROM_IMAGES = Vec3Int(128, 128, 1)
DEFAULT_BIT_DEPTH = 8
PROPERTIES_FILE_NAME = "datasource-properties.json"
ZGROUP_FILE_NAME = ".zgroup"
ZATTRS_FILE_NAME = ".zattrs"
ZARR_JSON_FILE_NAME = "zarr.json"



import json
import re
from collections import defaultdict
from typing import Dict, Generator, Iterable, List, Optional, Tuple, Union, cast

import attr
import numpy as np

from .mag import Mag
from .vec3_int import Vec3Int, Vec3IntLike

_DEFAULT_BBOX_NAME = "Unnamed Bounding Box"


@attr.frozen
class BoundingBox:
    """
    This class is used to represent an axis-aligned cuboid in 3D.
    The top-left coordinate is inclusive and the bottom-right coordinate is exclusive.

    A small usage example:

    ```python
    from webknossos import BoundingBox

    bbox_1 = BoundingBox((0, 0, 0), (100, 100, 100))
    bbox_2 = BoundingBox((75, 75, 75), (100, 100, 100))

    assert bbox_1.intersected_with(bbox_2).size == (25, 25, 25)
    ```
    """

    topleft: Vec3Int = attr.field(converter=Vec3Int)
    size: Vec3Int = attr.field(converter=Vec3Int)
    bottomright: Vec3Int = attr.field(init=False)
    name: Optional[str] = _DEFAULT_BBOX_NAME
    is_visible: bool = True
    color: Optional[Tuple[float, float, float, float]] = None

    def __attrs_post_init__(self) -> None:
        if not self.size.is_positive():
            # Flip the size in negative dimensions, so that the topleft is smaller than bottomright.
            # E.g. BoundingBox((10, 10, 10), (-5, 5, 5)) -> BoundingBox((5, 10, 10), (5, 5, 5)).
            negative_size = self.size.pairmin(Vec3Int.zeros())
            new_topleft = self.topleft + negative_size
            new_size = self.size.pairmax(-self.size)
            object.__setattr__(self, "topleft", new_topleft)
            object.__setattr__(self, "size", new_size)

        # Compute bottomright to avoid that it's recomputed every time
        # it is needed.
        object.__setattr__(self, "bottomright", self.topleft + self.size)

    def with_topleft(self, new_topleft: Vec3IntLike) -> "BoundingBox":
        return attr.evolve(self, topleft=new_topleft)

    def with_size(self, new_size: Vec3IntLike) -> "BoundingBox":
        return attr.evolve(self, size=new_size)

    def with_name(self, name: Optional[str]) -> "BoundingBox":
        return attr.evolve(self, name=name)

    def with_is_visible(self, is_visible: bool) -> "BoundingBox":
        return attr.evolve(self, is_visible=is_visible)

    def with_color(
        self, color: Optional[Tuple[float, float, float, float]]
    ) -> "BoundingBox":
        return attr.evolve(self, color=color)

    def with_bounds_x(
        self, new_topleft_x: Optional[int] = None, new_size_x: Optional[int] = None
    ) -> "BoundingBox":
        """Returns a copy of the bounding box with topleft.x optionally replaced and size.x optionally replaced."""

        new_topleft = (
            self.topleft.with_x(new_topleft_x)
            if new_topleft_x is not None
            else self.topleft
        )
        new_size = self.size.with_x(new_size_x) if new_size_x is not None else self.size
        return attr.evolve(self, topleft=new_topleft, size=new_size)

    def with_bounds_y(
        self, new_topleft_y: Optional[int] = None, new_size_y: Optional[int] = None
    ) -> "BoundingBox":
        """Returns a copy of the bounding box with topleft.y optionally replaced and size.y optionally replaced."""

        new_topleft = (
            self.topleft.with_y(new_topleft_y)
            if new_topleft_y is not None
            else self.topleft
        )
        new_size = self.size.with_y(new_size_y) if new_size_y is not None else self.size
        return attr.evolve(self, topleft=new_topleft, size=new_size)

    def with_bounds_z(
        self, new_topleft_z: Optional[int] = None, new_size_z: Optional[int] = None
    ) -> "BoundingBox":
        """Returns a copy of the bounding box with topleft.z optionally replaced and size.z optionally replaced."""

        new_topleft = (
            self.topleft.with_z(new_topleft_z)
            if new_topleft_z is not None
            else self.topleft
        )
        new_size = self.size.with_z(new_size_z) if new_size_z is not None else self.size
        return attr.evolve(self, topleft=new_topleft, size=new_size)

    @classmethod
    def from_wkw_dict(cls, bbox: Dict) -> "BoundingBox":
        return cls(bbox["topLeft"], [bbox["width"], bbox["height"], bbox["depth"]])

    @classmethod
    def from_config_dict(cls, bbox: Dict) -> "BoundingBox":
        return cls(bbox["topleft"], bbox["size"])

    @classmethod
    def from_tuple6(cls, tuple6: Tuple[int, int, int, int, int, int]) -> "BoundingBox":
        return cls(tuple6[0:3], tuple6[3:6])

    @classmethod
    def from_tuple2(cls, tuple2: Tuple[Vec3IntLike, Vec3IntLike]) -> "BoundingBox":
        return cls(tuple2[0], tuple2[1])

    @classmethod
    def from_points(cls, points: Iterable[Vec3IntLike]) -> "BoundingBox":
        """Returns a bounding box exactly containing all points."""

        all_points = np.array([Vec3Int(point).to_list() for point in points])
        topleft = all_points.min(axis=0)
        bottomright = all_points.max(axis=0)

        # bottomright is exclusive
        bottomright += 1

        return cls(topleft, bottomright - topleft)

    @classmethod
    def from_checkpoint_name(cls, checkpoint_name: str) -> "BoundingBox":
        """This function extracts a bounding box in the format `x_y_z_sx_sy_xz` which is contained in a string."""
        regex = r"(([0-9]+_){5}([0-9]+))"
        match = re.search(regex, checkpoint_name)
        assert (
            match is not None
        ), f"Could not extract bounding box from {checkpoint_name}"
        bbox_tuple = tuple(int(value) for value in match.group().split("_"))
        return cls.from_tuple6(cast(Tuple[int, int, int, int, int, int], bbox_tuple))

    @classmethod
    def from_csv(cls, csv_bbox: str) -> "BoundingBox":
        bbox_tuple = tuple(int(x) for x in csv_bbox.split(","))
        return cls.from_tuple6(cast(Tuple[int, int, int, int, int, int], bbox_tuple))

    @classmethod
    def from_auto(
        cls, obj: Union["BoundingBox", str, Dict, List, Tuple]
    ) -> "BoundingBox":
        if isinstance(obj, cls):
            return obj
        elif isinstance(obj, str):
            if ":" in obj:
                return cls.from_auto(json.loads(obj))
            else:
                return cls.from_csv(obj)
        elif isinstance(obj, dict):
            if "size" in obj:
                return cls.from_config_dict(obj)
            return cls.from_wkw_dict(obj)
        elif isinstance(obj, list) or isinstance(obj, tuple):
            if len(obj) == 2:
                return cls.from_tuple2(obj)  # type: ignore
            elif len(obj) == 6:
                return cls.from_tuple6(obj)  # type: ignore

        raise Exception("Unknown bounding box format.")

    @classmethod
    def group_boxes_with_aligned_mag(
        cls, bounding_boxes: Iterable["BoundingBox"], aligning_mag: Mag
    ) -> Dict["BoundingBox", List["BoundingBox"]]:
        """
        Groups the given BoundingBox instances by aligning each
        bbox to the given mag and using that as the key.
        For example, bounding boxes of size 256**3 could be grouped
        into the corresponding 1024**3 chunks to which they belong
        by using aligning_mag = Mag(1024).
        """

        chunks_with_bboxes = defaultdict(list)
        for bbox in bounding_boxes:
            chunk_key = bbox.align_with_mag(aligning_mag, ceil=True)
            chunks_with_bboxes[chunk_key].append(bbox)

        return chunks_with_bboxes

    @classmethod
    def empty(
        cls,
    ) -> "BoundingBox":
        return cls(Vec3Int.zeros(), Vec3Int.zeros())

    def to_wkw_dict(self) -> dict:
        (  # pylint: disable=unbalanced-tuple-unpacking
            width,
            height,
            depth,
        ) = self.size.to_list()

        return {
            "topLeft": self.topleft.to_list(),
            "width": width,
            "height": height,
            "depth": depth,
        }

    def to_config_dict(self) -> dict:
        return {"topleft": self.topleft.to_list(), "size": self.size.to_list()}

    def to_checkpoint_name(self) -> str:
        x, y, z = self.topleft
        width, height, depth = self.size
        return "{x}_{y}_{z}_{width}_{height}_{depth}".format(
            x=x, y=y, z=z, width=width, height=height, depth=depth
        )

    def to_tuple6(self) -> Tuple[int, int, int, int, int, int]:
        return tuple(self.topleft.to_list() + self.size.to_list())  # type: ignore

    def to_csv(self) -> str:
        return ",".join(map(str, self.to_tuple6()))

    def __repr__(self) -> str:
        return "BoundingBox(topleft={}, size={})".format(
            str(tuple(self.topleft)), str(tuple(self.size))
        )

    def __str__(self) -> str:
        return self.__repr__()

    def __eq__(self, other: object) -> bool:
        if isinstance(other, BoundingBox):
            return self.topleft == other.topleft and self.size == other.size
        else:
            raise NotImplementedError()

    def padded_with_margins(
        self, margins_left: Vec3IntLike, margins_right: Optional[Vec3IntLike] = None
    ) -> "BoundingBox":
        if margins_right is None:
            margins_right = margins_left

        margins_left = Vec3Int(margins_left)
        margins_right = Vec3Int(margins_right)

        return attr.evolve(
            self,
            topleft=self.topleft - margins_left,
            size=self.size + (margins_left + margins_right),
        )

    def intersected_with(
        self, other: "BoundingBox", dont_assert: bool = False
    ) -> "BoundingBox":
        """If dont_assert is set to False, this method may return empty bounding boxes (size == (0, 0, 0))"""

        topleft = self.topleft.pairmax(other.topleft)
        bottomright = self.bottomright.pairmin(other.bottomright)
        size = (bottomright - topleft).pairmax(Vec3Int.zeros())

        intersection = attr.evolve(self, topleft=topleft, size=size)

        if not dont_assert:
            assert (
                not intersection.is_empty()
            ), f"No intersection between bounding boxes {self} and {other}."

        return intersection

    def extended_by(self, other: "BoundingBox") -> "BoundingBox":
        if self.is_empty():
            return other
        if other.is_empty():
            return self

        topleft = self.topleft.pairmin(other.topleft)
        bottomright = self.bottomright.pairmax(other.bottomright)
        size = bottomright - topleft

        return attr.evolve(self, topleft=topleft, size=size)

    def is_empty(self) -> bool:
        return not self.size.is_positive(strictly_positive=True)

    def in_mag(self, mag: Mag) -> "BoundingBox":
        mag_vec = mag.to_vec3_int()

        assert (
            self.topleft % mag_vec == Vec3Int.zeros()
        ), f"topleft {self.topleft} is not aligned with the mag {mag}. Use BoundingBox.align_with_mag()."
        assert (
            self.bottomright % mag_vec == Vec3Int.zeros()
        ), f"bottomright {self.bottomright} is not aligned with the mag {mag}. Use BoundingBox.align_with_mag()."

        return attr.evolve(
            self,
            topleft=(self.topleft // mag_vec),
            size=(self.size // mag_vec),
        )

    def from_mag_to_mag1(self, from_mag: Mag) -> "BoundingBox":
        mag_vec = from_mag.to_vec3_int()
        return attr.evolve(
            self,
            topleft=(self.topleft * mag_vec),
            size=(self.size * mag_vec),
        )

    def _align_with_mag_slow(self, mag: Mag, ceil: bool = False) -> "BoundingBox":
        """Rounds the bounding box, so that both topleft and bottomright are divisible by mag.

        :argument ceil: If true, the bounding box is enlarged when necessary. If false, it's shrinked when necessary.
        """
        np_mag = mag.to_np()

        align = (
            lambda point, round_fn: round_fn(point.to_np() / np_mag).astype(int)
            * np_mag
        )

        if ceil:
            topleft = align(self.topleft, np.floor)
            bottomright = align(self.bottomright, np.ceil)
        else:
            topleft = align(self.topleft, np.ceil)
            bottomright = align(self.bottomright, np.floor)
        return attr.evolve(self, topleft=topleft, size=bottomright - topleft)

    def align_with_mag(
        self, mag: Union[Mag, Vec3Int], ceil: bool = False
    ) -> "BoundingBox":
        """Rounds the bounding box, so that both topleft and bottomright are divisible by mag.

        :argument ceil: If true, the bounding box is enlarged when necessary. If false, it's shrinked when necessary.
        """
        # This does the same as _align_with_mag_slow, which is more readable.
        # Same behavior is asserted in test_align_with_mag_against_numpy_implementation
        mag_vec = mag.to_vec3_int() if isinstance(mag, Mag) else mag
        roundup = self.topleft if ceil else self.bottomright
        rounddown = self.bottomright if ceil else self.topleft
        margin_to_roundup = roundup % mag_vec
        aligned_roundup = roundup - margin_to_roundup
        margin_to_rounddown = (mag_vec - (rounddown % mag_vec)) % mag_vec
        aligned_rounddown = rounddown + margin_to_rounddown
        if ceil:
            return attr.evolve(
                self, topleft=aligned_roundup, size=aligned_rounddown - aligned_roundup
            )
        else:
            return attr.evolve(
                self,
                topleft=aligned_rounddown,
                size=aligned_roundup - aligned_rounddown,
            )

    def contains(self, coord: Union[Vec3IntLike, np.ndarray]) -> bool:
        """Check whether a point is inside of the bounding box.
        Note that the point may have float coordinates in the ndarray case"""

        if isinstance(coord, np.ndarray):
            assert coord.shape == (
                3,
            ), f"Numpy array BoundingBox.contains must have shape (3,), got {coord.shape}."
            return cast(
                bool,
                np.all(coord >= self.topleft) and np.all(coord < self.bottomright),
            )
        else:
            # In earlier versions, we simply converted to ndarray to have
            # a unified calculation here, but this turned out to be a performance bottleneck.
            # Therefore, the contains-check is performed on the tuple here.
            coord = Vec3Int(coord)
            return (
                self.topleft[0] <= coord[0] < self.bottomright[0]
                and self.topleft[1] <= coord[1] < self.bottomright[1]
                and self.topleft[2] <= coord[2] < self.bottomright[2]
            )

    def contains_bbox(self, inner_bbox: "BoundingBox") -> bool:
        return inner_bbox.intersected_with(self, dont_assert=True) == inner_bbox

    def chunk(
        self,
        chunk_shape: Vec3IntLike,
        chunk_border_alignments: Optional[Vec3IntLike] = None,
    ) -> Generator["BoundingBox", None, None]:
        """Decompose the bounding box into smaller chunks of size `chunk_shape`.

        Chunks at the border of the bounding box might be smaller than chunk_shape.
        If `chunk_border_alignment` is set, all border coordinates
        *between two chunks* will be divisible by that value.
        """

        start = self.topleft.to_np()
        chunk_shape = Vec3Int(chunk_shape).to_np()

        start_adjust = np.array([0, 0, 0])
        if chunk_border_alignments is not None:
            chunk_border_alignments_array = Vec3Int(chunk_border_alignments).to_np()
            assert np.all(
                chunk_shape % chunk_border_alignments_array == 0
            ), f"{chunk_shape} not divisible by {chunk_border_alignments_array}"

            # Move the start to be aligned correctly. This doesn't actually change
            # the start of the first chunk, because we'll intersect with `self`,
            # but it'll lead to all chunk borders being aligned correctly.
            start_adjust = start % chunk_border_alignments_array

        for x in range(
            start[0] - start_adjust[0], start[0] + self.size[0], chunk_shape[0]
        ):
            for y in range(
                start[1] - start_adjust[1], start[1] + self.size[1], chunk_shape[1]
            ):
                for z in range(
                    start[2] - start_adjust[2], start[2] + self.size[2], chunk_shape[2]
                ):
                    yield BoundingBox([x, y, z], chunk_shape).intersected_with(self)

    def volume(self) -> int:
        return self.size.prod()

    def slice_array(self, array: np.ndarray) -> np.ndarray:
        return array[
            self.topleft.x : self.bottomright.x,
            self.topleft.y : self.bottomright.y,
            self.topleft.z : self.bottomright.z,
        ]

    def to_slices(self) -> Tuple[slice, slice, slice]:
        return np.index_exp[
            self.topleft.x : self.bottomright.x,
            self.topleft.y : self.bottomright.y,
            self.topleft.z : self.bottomright.z,
        ]

    def offset(self, vector: Vec3IntLike) -> "BoundingBox":
        return attr.evolve(self, topleft=self.topleft + Vec3Int(vector))

    def __hash__(self) -> int:
        return hash(self.to_tuple6())



import logging
import math
from itertools import product
from typing import List, Tuple

import numpy as np

from ..geometry import BoundingBox, Vec3Int
from .view import View


def _vec3int_mulf(vec: Vec3Int, factors: List[float]) -> Vec3Int:
    return Vec3Int(
        int(vec.x * factors[0]), int(vec.y * factors[1]), int(vec.z * factors[2])
    )


def upsample_cube(cube_buffer: np.ndarray, factors: List[int]) -> np.ndarray:
    ds = cube_buffer.shape
    out_buf = np.zeros(tuple(s * f for s, f in zip(ds, factors)), cube_buffer.dtype)
    for dx in (0, factors[0] - 1):
        for dy in (0, factors[1] - 1):
            for dz in (0, factors[2] - 1):
                out_buf[
                    dx : out_buf.shape[0] : factors[0],
                    dy : out_buf.shape[1] : factors[1],
                    dz : out_buf.shape[2] : factors[2],
                ] = cube_buffer
    return out_buf


def upsample_cube_job(
    args: Tuple[View, View, int],
    mag_factors: List[float],
    buffer_shape: Vec3Int,
) -> None:
    (source_view, target_view, _i) = args

    assert all(
        1 >= f for f in mag_factors
    ), f"mag_factors ({mag_factors}) for upsampling must be smaller than 1"

    try:
        num_channels = target_view.info.num_channels
        target_size = target_view.bounding_box.in_mag(target_view.mag).size
        shape = (num_channels,) + target_size.to_tuple()
        file_buffer = np.zeros(shape, target_view.get_dtype())

        tiles = product(
            *list(
                [
                    list(range(0, math.ceil(len)))
                    for len in target_size.to_np() / buffer_shape.to_np()
                ]
            )
        )

        for tile in tiles:
            target_offset = Vec3Int(tile) * buffer_shape
            source_offset = _vec3int_mulf(target_offset, mag_factors)
            source_size = source_view.bounding_box.in_mag(source_view.mag).size
            source_size = _vec3int_mulf(buffer_shape, mag_factors).pairmin(
                source_size - source_offset
            )

            bbox = BoundingBox(source_offset, source_size)
            cube_buffer_channels = source_view.read(
                relative_bounding_box=bbox.from_mag_to_mag1(source_view.mag),
            )

            for channel_index in range(num_channels):
                cube_buffer = cube_buffer_channels[channel_index]

                if not np.all(cube_buffer == 0):
                    # Upsample the buffer
                    inverse_factors = [int(1 / f) for f in mag_factors]
                    data_cube = upsample_cube(cube_buffer, inverse_factors)

                    buffer_offset = target_offset
                    buffer_end = buffer_offset + data_cube.shape

                    file_buffer[
                        channel_index,
                        buffer_offset[0] : buffer_end[0],
                        buffer_offset[1] : buffer_end[1],
                        buffer_offset[2] : buffer_end[2],
                    ] = data_cube

        # Write the upsampled buffer to target
        if source_view.info.num_channels == 1:
            file_buffer = file_buffer[0]  # remove channel dimension
        target_view.write(file_buffer)

    except Exception as exc:
        logging.error(
            f"Upsampling of target {target_view.bounding_box} failed with {exc}"
        )
        raise exc



import math
import re
from functools import total_ordering
from math import log2
from typing import Any, Iterator, List, Optional, Tuple, cast

import attr
import numpy as np

from .vec3_int import Vec3Int, Vec3IntLike


def _import_mag(mag_like: Any) -> Vec3Int:
    as_vec3_int: Optional[Vec3Int] = None

    if isinstance(mag_like, Mag):
        as_vec3_int = mag_like.to_vec3_int()
    elif isinstance(mag_like, int):
        as_vec3_int = Vec3Int(mag_like, mag_like, mag_like)
    elif isinstance(mag_like, Vec3Int):
        as_vec3_int = mag_like
    elif isinstance(mag_like, list) or isinstance(mag_like, tuple):
        as_vec3_int = Vec3Int(cast(Vec3IntLike, mag_like))
    elif isinstance(mag_like, str):
        if re.match(r"^\d+$", mag_like) is not None:
            as_vec3_int = Vec3Int(int(mag_like), int(mag_like), int(mag_like))
        elif re.match(r"^\d+-\d+-\d+$", mag_like) is not None:
            as_vec3_int = Vec3Int([int(m) for m in mag_like.split("-")])
    elif isinstance(mag_like, np.ndarray):
        as_vec3_int = Vec3Int(mag_like)

    if as_vec3_int is None:
        raise ValueError(
            f"Mag must be int or a vector3 of ints or a string shaped like e.g. 2-2-1. Got: {mag_like}"
        )
    for m in as_vec3_int:
        assert (
            log2(m) % 1 == 0
        ), f"Mag components must be power of 2, got {m} in {as_vec3_int}."

    return as_vec3_int


@total_ordering
@attr.frozen(order=False)
class Mag:
    """
    Represents the magnification level of a data layer. For example, the finest
    quality is usally not downsampled and is represented by Mag(1).
    When data is downsampled by a factor of 4 in all dimensions, this is referred
    to as Mag(4).
    When data is downsampled anisotropically by 2 in x and y and not downsampled in
    z, this is referred to as Mag(2, 2, 1).
    """

    _mag: Vec3Int = attr.ib(converter=_import_mag)

    @property
    def x(self) -> int:
        return self._mag.x

    @property
    def y(self) -> int:
        return self._mag.y

    @property
    def z(self) -> int:
        return self._mag.z

    @property
    def max_dim(self) -> int:
        return max(self._mag)

    @property
    def max_dim_log2(self) -> int:
        return int(math.log(self.max_dim) / math.log(2))

    def __lt__(self, other: Any) -> bool:
        return self.max_dim < Mag(other).max_dim

    def __le__(self, other: Any) -> bool:
        return self.max_dim <= Mag(other).max_dim

    def __eq__(self, other: Any) -> bool:
        return self.to_vec3_int() == Mag(other).to_vec3_int()

    def __str__(self) -> str:
        return self.to_layer_name()

    def __repr__(self) -> str:
        return f"Mag({self.to_layer_name()})"

    def to_layer_name(self) -> str:
        x, y, z = self._mag
        if x == y and y == z:
            return str(x)
        else:
            return self.to_long_layer_name()

    def to_long_layer_name(self) -> str:
        x, y, z = self._mag
        return "{}-{}-{}".format(x, y, z)

    def to_list(self) -> List[int]:
        return self._mag.to_list()

    def to_np(self) -> np.ndarray:
        return self._mag.to_np()

    def to_vec3_int(self) -> Vec3Int:
        return self._mag

    def to_tuple(self) -> Tuple[int, int, int]:
        return self._mag.to_tuple()

    def __mul__(self, factor: int) -> "Mag":
        return Mag(self._mag * factor)

    def __floordiv__(self, d: int) -> "Mag":
        return Mag(self._mag // d)

    def __hash__(self) -> int:
        return hash(self._mag)

    def __iter__(self) -> Iterator[int]:
        return iter(self._mag)



from os import PathLike
from pathlib import Path
from typing import Set

import numpy as np

try:
    from pims import FramesSequenceND
except ImportError as e:
    raise RuntimeError(
        "Cannot import pims, please install it e.g. using 'webknossos[all]'"
    ) from e

try:
    import tifffile
except ImportError as e:
    raise RuntimeError(
        "Cannot import tifffile, please install it e.g. using 'webknossos[tifffile]'"
    ) from e


class PimsImagejTiffReader(FramesSequenceND):
    @classmethod
    def class_exts(cls) -> Set[str]:
        return {"tif", "tiff"}

    # class_priority is used in pims to pick the reader with the highest priority.
    # Default is 10, and bioformats priority is 2.
    # See http://soft-matter.github.io/pims/v0.6.1/custom_readers.html#plugging-into-pims-s-open-function
    class_priority = 20

    def __init__(self, path: PathLike) -> None:
        super().__init__()
        path = Path(path)
        tiff = tifffile.TiffFile(path)
        assert tiff.is_imagej, f"{path} is not an ImageJ Tiff"
        channels = tiff.imagej_metadata["channels"]
        z = tiff.imagej_metadata["images"] / channels

        self.memmap = tifffile.memmap(path)
        # shape should be zcyx
        assert len(self.memmap.shape) == 4
        assert self.memmap.shape[0] == z
        assert self.memmap.shape[1] == channels

        self._init_axis("x", self.memmap.shape[3])
        self._init_axis("y", self.memmap.shape[2])
        self._init_axis("z", z)
        self._init_axis("c", channels)
        self._register_get_frame(self._get_frame, "cyx")

    @property
    def pixel_type(self) -> np.dtype:
        return self.memmap.dtype

    def _get_frame(self, **ind: int) -> np.ndarray:
        return self.memmap[ind["z"]]



import xml.etree.ElementTree as ET
from typing import IO, BinaryIO, List, NamedTuple, Optional

from loxun import XmlWriter

from .branchpoint import Branchpoint
from .comment import Comment
from .edge import Edge
from .group import Group
from .meta import Meta
from .node import Node
from .parameters import Parameters
from .segment import Segment
from .tree import Tree
from .volume import Volume


class Nml(NamedTuple):
    meta: List[Meta]
    parameters: Parameters  # metadata for annotations
    trees: List[Tree]
    branchpoints: List[Branchpoint]
    comments: List[Comment]
    groups: List[Group]
    volumes: List[
        Volume
    ] = []  # reference to any volume data that is part of this annotation

    def get_meta(self, key: str) -> Optional[str]:
        for entry in self.meta:
            if entry.name == key:
                return entry.content
        return None

    def write(self, file: BinaryIO) -> None:
        """
        Writes the NML object to a file on disk.

        Arguments:
            file (BinaryIO): A Python file handle

        Example:
            ```
            with open("out.nml", "wb") as f:
                nml.write(f, nml)
            ```
        """
        with XmlWriter(file) as xf:
            self._dump(xf)

    def _dump(self, xf: XmlWriter) -> None:
        xf.startTag("things")

        self.parameters._dump(xf)

        for meta_i in self.meta:
            meta_i._dump(xf)

        for tree in self.trees:
            tree._dump(xf)

        xf.startTag("branchpoints")
        for branchpoint in self.branchpoints:
            branchpoint._dump(xf)
        xf.endTag()  # branchpoints

        xf.startTag("comments")
        for comment in self.comments:
            comment._dump(xf)
        xf.endTag()  # comments

        xf.startTag("groups")
        for group in self.groups:
            group._dump(xf)
        xf.endTag()  # groups

        for volume in self.volumes:
            volume._dump(xf)

        xf.endTag()  # things

    @classmethod
    def parse(cls, file: IO[bytes]) -> "Nml":
        """
        Reads a WEBKNOSSOS NML skeleton file from disk, _parses it and returns an NML Python object

        Arguments:
            file (BinaryIO): A Python file handle

        Return:
            NML: A WEBKNOSSOS skeleton annotation as Python NML object

        Example:
            ```
            with open("input.nml", "rb") as f:
                nml = Nml.parse(f)
            ```
        """

        meta = []
        parameters = None
        trees = []
        branchpoints = []
        comments = []
        current_tree = None
        root_group = Group(-1, "", [])
        group_stack = [root_group]
        element_stack = []
        volumes = []

        for event, elem in ET.iterparse(file, events=("start", "end")):
            if event == "start":
                element_stack.append(elem)
                if elem.tag == "thing":
                    current_tree = Tree._parse(elem)
                    trees.append(current_tree)
                elif elem.tag == "node":
                    assert (
                        current_tree is not None
                    ), "<node ...> tag needs to be child of a <thing ...> tag."
                    current_tree.nodes.append(Node._parse(elem))
                elif elem.tag == "edge":
                    assert (
                        current_tree is not None
                    ), "<edge ...> tag needs to be child of a <thing ...> tag."
                    current_tree.edges.append(Edge._parse(elem))
                elif elem.tag == "branchpoint":
                    branchpoints.append(Branchpoint._parse(elem))
                elif elem.tag == "comment":
                    comments.append(Comment._parse(elem))
                elif elem.tag == "volume":
                    volumes.append(Volume._parse(elem))
                elif elem.tag == "meta":
                    meta.append(Meta._parse(elem))
                elif elem.tag == "group":
                    group = Group._parse(elem)
                    group_stack[-1].children.append(group)
                    group_stack.append(group)
                elif elem.tag == "segment":
                    segment = Segment._parse(elem)
                    if volumes[-1].segments is None:
                        volumes[-1].segments = []
                    volumes[-1].segments.append(segment)
            elif event == "end":
                if elem.tag == "parameters":
                    parameters = Parameters._parse(elem)
                elif elem.tag == "thing":
                    current_tree = None
                elif elem.tag == "group":
                    group_stack.pop()

                element_stack.pop()
                # Do not clear the elements of the parameters tag as we want to _parse those all at once
                # when the closing parameters tag is _parsed
                if len(element_stack) and element_stack[-1].tag != "parameters":
                    # Discard the element to save memory
                    elem.clear()

        assert parameters is not None, "No parameters found in NML"

        return cls(
            meta=meta,
            parameters=parameters,
            trees=trees,
            branchpoints=branchpoints,
            comments=comments,
            groups=root_group.children,
            volumes=volumes,
        )



from typing import NamedTuple
from xml.etree.ElementTree import Element

from loxun import XmlWriter

from .utils import enforce_not_null


class Comment(NamedTuple):
    node: int
    content: str  # free text field, supports Markdown formatting

    def _dump(self, xf: XmlWriter) -> None:
        xf.tag("comment", {"node": str(self.node), "content": self.content})

    @classmethod
    def _parse(cls, nml_comment: Element) -> "Comment":
        return cls(
            int(enforce_not_null(nml_comment.get("node"))),
            enforce_not_null(nml_comment.get("content")),
        )



from typing import Any, Dict, List, Optional, Tuple

import attr

# Request and response bodies for wk/datastore routes
# Should contain only the fields that are actually used by the python libs
# Note that keys are converted to/from camelCase to match this
# Optional fields in response bodies should always have “= None” defaults.
# When adding things here, check which fields are optional on the WK server side


@attr.s(auto_attribs=True)
class ApiWkBuildInfoWebknossos:
    name: str
    version: str


@attr.s(auto_attribs=True)
class ApiWkBuildInfo:
    webknossos: ApiWkBuildInfoWebknossos
    local_data_store_enabled: bool
    local_tracing_store_enabled: bool


@attr.s(auto_attribs=True)
class ApiShortLink:
    long_link: str


@attr.s(auto_attribs=True)
class ApiDataStore:
    url: str
    name: str
    allows_upload: bool


@attr.s(auto_attribs=True)
class ApiTeam:
    id: str
    name: str
    organization: str


@attr.s(auto_attribs=True)
class ApiBoundingBox:
    top_left: Tuple[int, int, int]
    width: int
    height: int
    depth: int


@attr.s(auto_attribs=True)
class ApiDataLayer:
    name: str
    category: str
    element_class: str
    bounding_box: ApiBoundingBox
    resolutions: List[Tuple[int, int, int]]
    largest_segment_id: Optional[int] = None
    default_view_configuration: Optional[Dict[str, Any]] = None


@attr.s(auto_attribs=True)
class ApiDataSource:
    data_layers: Optional[List[ApiDataLayer]] = None
    status: Optional[str] = None
    scale: Optional[Tuple[float, float, float]] = None


@attr.s(auto_attribs=True)
class ApiDataset:
    name: str
    is_public: bool
    folder_id: str
    allowed_teams: List[ApiTeam]
    tags: List[str]
    data_store: ApiDataStore
    data_source: ApiDataSource
    display_name: Optional[str] = None
    description: Optional[str] = None


@attr.s(auto_attribs=True)
class ApiSharingToken:
    sharing_token: str


@attr.s(auto_attribs=True)
class ApiDatasetUploadInformation:
    upload_id: str


@attr.s(auto_attribs=True)
class ApiLinkedLayerIdentifier:
    organization_name: str
    data_set_name: str
    layer_name: str
    new_layer_name: Optional[str] = None


@attr.s(auto_attribs=True)
class ApiReserveDatasetUploadInformation:
    upload_id: str
    name: str
    organization: str
    total_file_count: int
    initial_teams: List[str]
    layers_to_link: Optional[List[ApiLinkedLayerIdentifier]] = None
    folder_id: Optional[str] = None


@attr.s(auto_attribs=True)
class ApiTaskStatus:
    pending: int
    active: int
    finished: int


@attr.s(auto_attribs=True)
class ApiTaskType:
    id: str
    summary: str
    description: str
    team_id: str
    team_name: str


@attr.s(auto_attribs=True)
class ApiTask:
    id: str
    project_id: str
    data_set: str
    status: ApiTaskStatus
    type: ApiTaskType


@attr.s(auto_attribs=True)
class ApiExperience:
    domain: str
    value: int


@attr.s(auto_attribs=True)
class ApiNmlTaskParameters:
    task_type_id: str
    needed_experience: ApiExperience
    pending_instances: int
    project_name: str
    script_id: Optional[str]
    bounding_box: Optional[ApiBoundingBox]


@attr.s(auto_attribs=True)
class ApiTaskParameters:
    task_type_id: str
    needed_experience: ApiExperience
    pending_instances: int
    project_name: str
    script_id: Optional[str]
    bounding_box: Optional[ApiBoundingBox]
    data_set: str
    edit_position: Tuple[int, int, int]
    edit_rotation: Tuple[float, float, float]


@attr.s(auto_attribs=True)
class ApiSingleTaskCreationResult:
    error: Optional[str]
    success: Optional[ApiTask]


@attr.s(auto_attribs=True)
class ApiTaskCreationResult:
    tasks: List[ApiSingleTaskCreationResult]
    warnings: List[str]


@attr.s(auto_attribs=True)
class ApiTeamMembership:
    id: str
    name: str


@attr.s(auto_attribs=True)
class ApiUser:
    id: str
    email: str
    organization: str
    first_name: str
    last_name: str
    created: int
    last_activity: int
    is_active: bool
    is_admin: bool
    is_dataset_manager: bool
    teams: List[ApiTeamMembership]
    experiences: Dict[str, int]


@attr.s(auto_attribs=True)
class ApiUserCompact:
    id: str
    email: str
    first_name: str
    last_name: str
    is_admin: bool
    is_dataset_manager: bool


@attr.s(auto_attribs=True)
class ApiMonth:
    month: int
    year: int


@attr.s(auto_attribs=True)
class ApiLoggedTimeForMonth:
    payment_interval: ApiMonth
    duration_in_seconds: int


@attr.s(auto_attribs=True)
class ApiDataStoreToken:
    token: str


@attr.s(auto_attribs=True)
class ApiLoggedTimeGroupedByMonth:
    logged_time: List[ApiLoggedTimeForMonth]


@attr.s(auto_attribs=True)
class ApiProject:
    id: str
    name: str
    team: str
    team_name: str
    owner: ApiUser
    priority: int
    paused: bool
    expected_time: Optional[int] = None


@attr.s(auto_attribs=True)
class ApiAnnotation:
    id: str
    typ: str
    owner: ApiUserCompact
    name: str
    description: str
    state: str
    modified: int
    data_store: ApiDataStore
    tracing_time: Optional[int] = None  # millis


@attr.s(auto_attribs=True)
class ApiAnnotationIdentifier:
    id: str
    typ: str


@attr.s(auto_attribs=True)
class ApiAnnotationUploadResult:
    annotation: ApiAnnotationIdentifier
    messages: List[str]


@attr.s(auto_attribs=True)
class ApiFolderWithParent:
    id: str
    name: str
    parent: Optional[str] = None



"""This module takes care of downloading WEBKNOSSOS datasets."""

from typing import Any, List, Optional

import typer
from typing_extensions import Annotated

from ..annotation import Annotation
from ..client import webknossos_context
from ..dataset import Dataset
from ..geometry import BoundingBox, Mag
from ._utils import parse_bbox, parse_mag, parse_path


def main(
    *,
    target: Annotated[
        Any,
        typer.Argument(
            show_default=False,
            help="Path to save your WEBKNOSSOS dataset.",
            parser=parse_path,
        ),
    ],
    url: Annotated[
        str,
        typer.Option(
            help="URL of your dataset or your annotation.",
        ),
    ],
    token: Annotated[
        Optional[str],
        typer.Option(
            help="Authentication token for WEBKNOSSOS instance "
            "(https://webknossos.org/auth/token).",
            rich_help_panel="WEBKNOSSOS context",
            envvar="WK_TOKEN",
        ),
    ] = None,
    bbox: Annotated[
        Optional[BoundingBox],
        typer.Option(
            rich_help_panel="Partial download",
            help="Bounding box that should be downloaded. "
            "The input format is x,y,z,width,height,depth. "
            "Should be a comma seperated string (e.g. 0,0,0,10,10,10).",
            parser=parse_bbox,
            metavar="BBOX",
        ),
    ] = None,
    layer: Annotated[
        Optional[List[str]],
        typer.Option(
            rich_help_panel="Partial download",
            help="Layers that should be downloaded. "
            "For multiple layers type: --layer color --layer segmentation",
        ),
    ] = None,
    mag: Annotated[
        Optional[List[Mag]],
        typer.Option(
            rich_help_panel="Partial download",
            help="Mags that should be downloaded. "
            "Should be number or minus seperated string (e.g. 2 or 2-2-2). "
            "For multiple mags type: --mag 1 --mag 2",
            parser=parse_mag,
            metavar="MAG",
        ),
    ] = None,
) -> None:
    """Download a dataset from a WEBKNOSSOS server."""

    layers = layer if layer else None
    mags = mag if mag else None

    with webknossos_context(token=token):
        try:
            Dataset.download(
                dataset_name_or_url=url,
                path=target,
                bbox=bbox,
                layers=layers,
                mags=mags,
            )
        except AssertionError:
            Annotation.download(annotation_id_or_url=url).save(target)



import logging
import operator
import re
import warnings
from argparse import Namespace
from os import PathLike
from os.path import relpath
from pathlib import Path
from typing import TYPE_CHECKING, Dict, List, Optional, Tuple, Union

import numpy as np
from cluster_tools import Executor
from numpy.typing import DTypeLike
from upath import UPath

from ..geometry import BoundingBox, Mag, Vec3Int, Vec3IntLike
from ._array import ArrayException, BaseArray, DataFormat
from ._downsampling_utils import (
    calculate_default_coarsest_mag,
    calculate_mags_to_downsample,
    calculate_mags_to_upsample,
    determine_buffer_shape,
    downsample_cube_job,
    parse_interpolation_mode,
)
from ._upsampling_utils import upsample_cube_job
from .layer_categories import COLOR_CATEGORY, SEGMENTATION_CATEGORY, LayerCategoryType
from .properties import (
    LayerProperties,
    LayerViewConfiguration,
    MagViewProperties,
    SegmentationLayerProperties,
    _properties_floating_type_to_python_type,
    _python_floating_type_to_properties_type,
)
from .sampling_modes import SamplingModes
from .view import View, _copy_job

if TYPE_CHECKING:
    from .dataset import Dataset

from ..utils import (
    copytree,
    get_executor_for_args,
    is_fs_path,
    named_partial,
    rmtree,
    warn_deprecated,
)
from .defaults import (
    DEFAULT_CHUNK_SHAPE,
    DEFAULT_CHUNKS_PER_SHARD,
    DEFAULT_CHUNKS_PER_SHARD_ZARR,
)
from .mag_view import MagView, _find_mag_path_on_disk


def _is_int(s: str) -> bool:
    try:
        int(s)
        return True
    except ValueError:
        return False


def _convert_dtypes(
    dtype: DTypeLike,
    num_channels: int,
    dtype_per_layer_to_dtype_per_channel: bool,
) -> str:
    op = operator.truediv if dtype_per_layer_to_dtype_per_channel else operator.mul

    # split the dtype into the actual type and the number of bits
    # example: "uint24" -> ["uint", "24"]
    dtype_parts = re.split(r"(\d+)", str(dtype))
    # calculate number of bits for dtype_per_channel
    converted_dtype_parts = [
        (str(int(op(int(part), num_channels))) if _is_int(part) else part)
        for part in dtype_parts
    ]
    return "".join(converted_dtype_parts)


def _normalize_dtype_per_channel(dtype_per_channel: DTypeLike) -> np.dtype:
    try:
        return np.dtype(dtype_per_channel)
    except TypeError as e:
        raise TypeError(
            "Cannot add layer. The specified 'dtype_per_channel' must be a valid dtype."
        ) from e


def _normalize_dtype_per_layer(dtype_per_layer: DTypeLike) -> DTypeLike:
    try:
        dtype_per_layer = str(np.dtype(dtype_per_layer))
    except Exception:
        pass  # casting to np.dtype fails if the user specifies a special dtype like "uint24"
    return dtype_per_layer  # type: ignore[return-value]


def _dtype_per_layer_to_dtype_per_channel(
    dtype_per_layer: DTypeLike, num_channels: int
) -> np.dtype:
    try:
        return np.dtype(
            _convert_dtypes(
                dtype_per_layer, num_channels, dtype_per_layer_to_dtype_per_channel=True
            )
        )
    except TypeError as e:
        raise TypeError(
            "Converting dtype_per_layer to dtype_per_channel failed. Double check if the dtype_per_layer value is correct."
        ) from e


def _dtype_per_channel_to_dtype_per_layer(
    dtype_per_channel: DTypeLike, num_channels: int
) -> str:
    return _convert_dtypes(
        np.dtype(dtype_per_channel),
        num_channels,
        dtype_per_layer_to_dtype_per_channel=False,
    )


def _dtype_per_channel_to_element_class(
    dtype_per_channel: DTypeLike, num_channels: int
) -> str:
    dtype_per_layer = _dtype_per_channel_to_dtype_per_layer(
        dtype_per_channel, num_channels
    )
    return _python_floating_type_to_properties_type.get(
        dtype_per_layer, dtype_per_layer
    )


def _element_class_to_dtype_per_channel(
    element_class: str, num_channels: int
) -> np.dtype:
    dtype_per_layer = _properties_floating_type_to_python_type.get(
        element_class, element_class
    )
    return _dtype_per_layer_to_dtype_per_channel(dtype_per_layer, num_channels)


def _get_sharding_parameters(
    *,
    chunk_shape: Optional[Union[Vec3IntLike, int]],
    chunks_per_shard: Optional[Union[Vec3IntLike, int]],
    chunk_size: Optional[Union[Vec3IntLike, int]],  # deprecated
    block_len: Optional[int],  # deprecated
    file_len: Optional[int],  # deprecated
) -> Tuple[Optional[Vec3Int], Optional[Vec3Int]]:
    if chunk_shape is not None:
        chunk_shape = Vec3Int.from_vec_or_int(chunk_shape)
    elif chunk_size is not None:
        warn_deprecated("chunk_size", "chunk_shape")
        chunk_shape = Vec3Int.from_vec_or_int(chunk_size)
    elif block_len is not None:
        warn_deprecated("block_len", "chunk_shape")
        chunk_shape = Vec3Int.full(block_len)

    if chunks_per_shard is not None:
        chunks_per_shard = Vec3Int.from_vec_or_int(chunks_per_shard)
    elif file_len is not None:
        warn_deprecated("file_len", "chunks_per_shard")
        chunks_per_shard = Vec3Int.full(file_len)

    return (chunk_shape, chunks_per_shard)


class Layer:
    """
    A `Layer` consists of multiple `MagView`s, which store the same data in different magnifications.
    """

    def __init__(self, dataset: "Dataset", properties: LayerProperties) -> None:
        """
        Do not use this constructor manually. Instead use `Dataset``.add_layer()` to create a `Layer`.
        """
        # It is possible that the properties on disk do not contain the number of channels.
        # Therefore, the parameter is optional. However at this point, 'num_channels' was already inferred.
        assert properties.num_channels is not None

        self._name: str = (
            properties.name
        )  # The name is also stored in the properties, but the name is required to get the properties.
        self._dataset = dataset
        self._dtype_per_channel = _element_class_to_dtype_per_channel(
            properties.element_class, properties.num_channels
        )
        self._mags: Dict[Mag, MagView] = {}

        self.path.mkdir(parents=True, exist_ok=True)

        for mag in properties.mags:
            self._setup_mag(Mag(mag.mag))
        # Only keep the properties of mags that were initialized.
        # Sometimes the directory of a mag is removed from disk manually, but the properties are not updated.
        self._properties.mags = [
            res for res in self._properties.mags if Mag(res.mag) in self._mags
        ]

    @property
    def path(self) -> Path:
        return self.dataset.path / self.name

    @property
    def _properties(self) -> LayerProperties:
        return next(
            layer_property
            for layer_property in self.dataset._properties.data_layers
            if layer_property.name == self.name
        )

    @property
    def name(self) -> str:
        return self._name

    @name.setter
    def name(self, layer_name: str) -> None:
        """
        Renames the layer to `layer_name`. This changes the name of the directory on disk and updates the properties.
        Only layers on local file systems can be renamed.
        """
        if layer_name == self.name:
            return
        self.dataset._ensure_writable()
        assert (
            layer_name not in self.dataset.layers.keys()
        ), f"Failed to rename layer {self.name} to {layer_name}: The new name already exists."
        assert is_fs_path(self.path), f"Cannot rename remote layer {self.path}"
        assert (
            "/" not in layer_name
        ), f"Cannot rename layer, because there is a '/' character in the layer name: {layer_name}"
        self.path.rename(self.dataset.path / layer_name)
        del self.dataset.layers[self.name]
        self.dataset.layers[layer_name] = self
        self._properties.name = layer_name
        self._name = layer_name

        # The MagViews need to be updated
        for mag in self._mags.values():
            mag._path = _find_mag_path_on_disk(self.dataset.path, self.name, mag.name)
            # Deleting the dataset will close the file handle.
            # The new dataset will be opened automatically when needed.
            del mag._array

        self.dataset._export_as_json()

    @property
    def dataset(self) -> "Dataset":
        return self._dataset

    @property
    def bounding_box(self) -> BoundingBox:
        return self._properties.bounding_box

    @bounding_box.setter
    def bounding_box(self, bbox: BoundingBox) -> None:
        """
        Updates the offset and size of the bounding box of this layer in the properties.
        """
        self.dataset._ensure_writable()
        assert (
            bbox.topleft.is_positive()
        ), f"Updating the bounding box of layer {self} to {bbox} failed, topleft must not contain negative dimensions."
        self._properties.bounding_box = bbox
        self.dataset._export_as_json()
        for mag in self.mags.values():
            mag._array.ensure_size(
                bbox.align_with_mag(mag.mag).in_mag(mag.mag).bottomright
            )

    @property
    def category(self) -> LayerCategoryType:
        return COLOR_CATEGORY

    @property
    def dtype_per_layer(self) -> str:
        return _dtype_per_channel_to_dtype_per_layer(
            self.dtype_per_channel, self.num_channels
        )

    @property
    def dtype_per_channel(self) -> np.dtype:
        return self._dtype_per_channel

    @property
    def num_channels(self) -> int:
        assert self._properties.num_channels is not None
        return self._properties.num_channels

    @property
    def data_format(self) -> DataFormat:
        assert self._properties.data_format is not None
        return self._properties.data_format

    @property
    def default_view_configuration(self) -> Optional[LayerViewConfiguration]:
        return self._properties.default_view_configuration

    @default_view_configuration.setter
    def default_view_configuration(
        self, view_configuration: LayerViewConfiguration
    ) -> None:
        self.dataset._ensure_writable()
        self._properties.default_view_configuration = view_configuration
        self.dataset._export_as_json()  # update properties on disk

    @property
    def read_only(self) -> bool:
        return self.dataset.read_only

    @property
    def mags(self) -> Dict[Mag, MagView]:
        """
        Getter for dictionary containing all mags.
        """
        return self._mags

    def get_mag(self, mag: Union[int, str, list, tuple, np.ndarray, Mag]) -> MagView:
        """
        Returns the `MagView` called `mag` of this layer. The return type is `webknossos.dataset.mag_view.MagView`.

        This function raises an `IndexError` if the specified `mag` does not exist.
        """
        mag = Mag(mag)
        if mag not in self.mags.keys():
            raise IndexError(
                "The mag {} is not a mag of this layer".format(mag.to_layer_name())
            )
        return self.mags[mag]

    def get_finest_mag(self) -> MagView:
        return self.get_mag(min(self.mags.keys()))

    def get_best_mag(self) -> MagView:
        """Deprecated, please use `get_finest_mag`."""
        warn_deprecated("get_best_mag()", "get_finest_mag()")
        return self.get_finest_mag()

    def add_mag(
        self,
        mag: Union[int, str, list, tuple, np.ndarray, Mag],
        chunk_shape: Optional[Union[Vec3IntLike, int]] = None,  # DEFAULT_CHUNK_SHAPE,
        chunks_per_shard: Optional[
            Union[int, Vec3IntLike]
        ] = None,  # DEFAULT_CHUNKS_PER_SHARD,
        compress: bool = False,
        *,
        chunk_size: Optional[Union[Vec3IntLike, int]] = None,  # deprecated
        block_len: Optional[int] = None,  # deprecated
        file_len: Optional[int] = None,  # deprecated
    ) -> MagView:
        """
        Creates a new mag called and adds it to the layer.
        The parameter `chunk_shape`, `chunks_per_shard` and `compress` can be
        specified to adjust how the data is stored on disk.
        Note that writing compressed data which is not aligned with the blocks on disk may result in
        diminished performance, as full blocks will automatically be read to pad the write actions. Alternatively,
        you can call mag.compress() after all the data was written

        The return type is `webknossos.dataset.mag_view.MagView`.

        Raises an IndexError if the specified `mag` already exists.
        """
        self.dataset._ensure_writable()
        # normalize the name of the mag
        mag = Mag(mag)
        compression_mode = compress

        chunk_shape, chunks_per_shard = _get_sharding_parameters(
            chunk_shape=chunk_shape,
            chunks_per_shard=chunks_per_shard,
            chunk_size=chunk_size,
            block_len=block_len,
            file_len=file_len,
        )
        if chunk_shape is None:
            chunk_shape = DEFAULT_CHUNK_SHAPE
        if chunks_per_shard is None:
            if self.data_format == DataFormat.Zarr:
                chunks_per_shard = DEFAULT_CHUNKS_PER_SHARD_ZARR
            else:
                chunks_per_shard = DEFAULT_CHUNKS_PER_SHARD

        if chunk_shape not in (Vec3Int.full(32), Vec3Int.full(64)):
            warnings.warn(
                "[INFO] `chunk_shape` of `32, 32, 32` or `64, 64, 64` is recommended for optimal "
                + f"performance in WEBKNOSSOS. Got {chunk_shape}."
            )

        self._assert_mag_does_not_exist_yet(mag)
        self._create_dir_for_mag(mag)

        mag_view = MagView(
            self,
            mag,
            chunk_shape=chunk_shape,
            chunks_per_shard=chunks_per_shard,
            compression_mode=compression_mode,
            create=True,
        )

        mag_view._array.ensure_size(
            self.bounding_box.align_with_mag(mag).in_mag(mag).bottomright
        )

        self._mags[mag] = mag_view
        mag_array_info = mag_view.info
        self._properties.mags += [
            MagViewProperties(
                mag=Mag(mag_view.name),
                cube_length=(
                    mag_array_info.shard_shape.x
                    if mag_array_info.data_format == DataFormat.WKW
                    else None
                ),
                axis_order=(
                    {"x": 1, "y": 2, "z": 3, "c": 0}
                    if mag_array_info.data_format in (DataFormat.Zarr, DataFormat.Zarr3)
                    else None
                ),
            )
        ]

        self.dataset._export_as_json()

        return self._mags[mag]

    def add_mag_for_existing_files(
        self,
        mag: Union[int, str, list, tuple, np.ndarray, Mag],
    ) -> MagView:
        """
        Creates a new mag based on already existing files.

        Raises an IndexError if the specified `mag` does not exists.
        """
        self.dataset._ensure_writable()
        mag = Mag(mag)
        assert (
            mag not in self.mags
        ), f"Cannot add mag {mag} as it already exists for layer {self}"
        self._setup_mag(mag)
        mag_view = self._mags[mag]
        mag_array_info = mag_view.info
        self._properties.mags.append(
            MagViewProperties(
                mag=mag,
                cube_length=(
                    mag_array_info.shard_shape.x
                    if mag_array_info.data_format == DataFormat.WKW
                    else None
                ),
                axis_order=(
                    {"x": 1, "y": 2, "z": 3, "c": 0}
                    if mag_array_info.data_format in (DataFormat.Zarr, DataFormat.Zarr3)
                    else None
                ),
            )
        )
        self.dataset._export_as_json()

        return mag_view

    def get_or_add_mag(
        self,
        mag: Union[int, str, list, tuple, np.ndarray, Mag],
        chunk_shape: Optional[Union[Vec3IntLike, int]] = None,
        chunks_per_shard: Optional[Union[Vec3IntLike, int]] = None,
        compress: Optional[bool] = None,
        *,
        chunk_size: Optional[Union[Vec3IntLike, int]] = None,  # deprecated
        block_len: Optional[int] = None,  # deprecated
        file_len: Optional[int] = None,  # deprecated
    ) -> MagView:
        """
        Creates a new mag called and adds it to the dataset, in case it did not exist before.
        Then, returns the mag.

        See `add_mag` for more information.
        """

        # normalize the name of the mag
        mag = Mag(mag)
        compression_mode = compress

        chunk_shape, chunks_per_shard = _get_sharding_parameters(
            chunk_shape=chunk_shape,
            chunks_per_shard=chunks_per_shard,
            chunk_size=chunk_size,
            block_len=block_len,
            file_len=file_len,
        )

        if mag in self._mags.keys():
            assert (
                chunk_shape is None or self._mags[mag].info.chunk_shape == chunk_shape
            ), f"Cannot get_or_add_mag: The mag {mag} already exists, but the chunk sizes do not match"
            assert (
                chunks_per_shard is None
                or self._mags[mag].info.chunks_per_shard == chunks_per_shard
            ), f"Cannot get_or_add_mag: The mag {mag} already exists, but the chunks per shard do not match"
            assert (
                compression_mode is None
                or self._mags[mag].info.compression_mode == compression_mode
            ), f"Cannot get_or_add_mag: The mag {mag} already exists, but the compression modes do not match"
            return self.get_mag(mag)
        else:
            return self.add_mag(
                mag,
                chunk_shape=chunk_shape,
                chunks_per_shard=chunks_per_shard,
                compress=compression_mode or False,
            )

    def delete_mag(self, mag: Union[int, str, list, tuple, np.ndarray, Mag]) -> None:
        """
        Deletes the MagView from the `datasource-properties.json` and the data from disk.

        This function raises an `IndexError` if the specified `mag` does not exist.
        """
        self.dataset._ensure_writable()
        mag = Mag(mag)
        if mag not in self.mags.keys():
            raise IndexError(
                "Deleting mag {} failed. There is no mag with this name".format(mag)
            )

        del self._mags[mag]
        self._properties.mags = [
            res for res in self._properties.mags if Mag(res.mag) != mag
        ]
        self.dataset._export_as_json()
        # delete files on disk
        full_path = _find_mag_path_on_disk(
            self.dataset.path, self.name, mag.to_layer_name()
        )
        rmtree(full_path)

    def add_copy_mag(
        self,
        foreign_mag_view_or_path: Union[PathLike, str, MagView],
        extend_layer_bounding_box: bool = True,
        chunk_shape: Optional[Union[Vec3IntLike, int]] = None,
        chunks_per_shard: Optional[Union[Vec3IntLike, int]] = None,
        compress: Optional[bool] = None,
        executor: Optional[Executor] = None,
    ) -> MagView:
        """
        Copies the data at `foreign_mag_view_or_path` which can belong to another dataset
        to the current dataset. Additionally, the relevant information from the
        `datasource-properties.json` of the other dataset are copied, too.
        """
        self.dataset._ensure_writable()
        foreign_mag_view = MagView._ensure_mag_view(foreign_mag_view_or_path)
        self._assert_mag_does_not_exist_yet(foreign_mag_view.mag)

        mag_view = self.add_mag(
            mag=foreign_mag_view.mag,
            chunk_shape=chunk_shape or foreign_mag_view._array_info.chunk_shape,
            chunks_per_shard=chunks_per_shard
            or foreign_mag_view._array_info.chunks_per_shard,
            compress=compress
            if compress is not None
            else foreign_mag_view._array_info.compression_mode,
        )

        if extend_layer_bounding_box:
            self.bounding_box = self.bounding_box.extended_by(
                foreign_mag_view.layer.bounding_box
            )

        foreign_mag_view.for_zipped_chunks(
            func_per_chunk=_copy_job,
            target_view=mag_view,
            executor=executor,
            progress_desc=f"Copying mag {mag_view.mag.to_layer_name()} from {foreign_mag_view.layer} to {mag_view.layer}",
        )

        return mag_view

    def add_symlink_mag(
        self,
        foreign_mag_view_or_path: Union[PathLike, str, MagView],
        make_relative: bool = False,
        extend_layer_bounding_box: bool = True,
    ) -> MagView:
        """
        Creates a symlink to the data at `foreign_mag_view_or_path` which belongs to another dataset.
        The relevant information from the `datasource-properties.json` of the other dataset is copied to this dataset.
        Note: If the other dataset modifies its bounding box afterwards, the change does not affect this properties
        (or vice versa).
        If make_relative is True, the symlink is made relative to the current dataset path.
        Symlinked mags can only be added to layers on local file systems.
        """
        self.dataset._ensure_writable()
        foreign_mag_view = MagView._ensure_mag_view(foreign_mag_view_or_path)
        self._assert_mag_does_not_exist_yet(foreign_mag_view.mag)

        assert is_fs_path(
            self.path
        ), f"Cannot create symlinks in remote layer {self.path}"
        assert is_fs_path(
            foreign_mag_view.path
        ), f"Cannot create symlink to remote mag {foreign_mag_view.path}"

        foreign_normalized_mag_path = (
            Path(relpath(foreign_mag_view.path, self.path))
            if make_relative
            else foreign_mag_view.path.resolve()
        )

        (self.path / str(foreign_mag_view.mag)).symlink_to(foreign_normalized_mag_path)

        mag = self.add_mag_for_existing_files(foreign_mag_view.mag)

        if extend_layer_bounding_box:
            self.bounding_box = self.bounding_box.extended_by(
                foreign_mag_view.layer.bounding_box
            )
        return mag

    def add_fs_copy_mag(
        self,
        foreign_mag_view_or_path: Union[PathLike, str, MagView],
        extend_layer_bounding_box: bool = True,
    ) -> MagView:
        """
        Copies the data at `foreign_mag_view_or_path` which belongs to another dataset to the current dataset via the filesystem.
        Additionally, the relevant information from the `datasource-properties.json` of the other dataset are copied, too.
        """
        self.dataset._ensure_writable()
        foreign_mag_view = MagView._ensure_mag_view(foreign_mag_view_or_path)
        self._assert_mag_does_not_exist_yet(foreign_mag_view.mag)

        copytree(
            foreign_mag_view.path,
            self.path / str(foreign_mag_view.mag),
        )

        mag = self.add_mag_for_existing_files(foreign_mag_view.mag)

        if extend_layer_bounding_box:
            self.bounding_box = self.bounding_box.extended_by(
                foreign_mag_view.layer.bounding_box
            )

        return mag

    def _create_dir_for_mag(
        self, mag: Union[int, str, list, tuple, np.ndarray, Mag]
    ) -> None:
        mag = Mag(mag).to_layer_name()
        full_path = self.path / mag
        full_path.mkdir(parents=True, exist_ok=True)

    def _assert_mag_does_not_exist_yet(
        self, mag: Union[int, str, list, tuple, np.ndarray, Mag]
    ) -> None:
        if mag in self.mags.keys():
            raise IndexError(
                "Adding mag {} failed. There is already a mag with this name".format(
                    mag
                )
            )

    def _get_dataset_from_align_with_other_layers(
        self, align_with_other_layers: Union[bool, "Dataset"]
    ) -> Optional["Dataset"]:
        if isinstance(align_with_other_layers, bool):
            return self.dataset if align_with_other_layers else None
        else:
            return align_with_other_layers

    def downsample(
        self,
        from_mag: Optional[Mag] = None,
        coarsest_mag: Optional[Mag] = None,
        interpolation_mode: str = "default",
        compress: bool = True,
        sampling_mode: Union[str, SamplingModes] = SamplingModes.ANISOTROPIC,
        align_with_other_layers: Union[bool, "Dataset"] = True,
        buffer_shape: Optional[Vec3Int] = None,
        force_sampling_scheme: bool = False,
        args: Optional[Namespace] = None,  # deprecated
        allow_overwrite: bool = False,
        only_setup_mags: bool = False,
        executor: Optional[Executor] = None,
    ) -> None:
        """
        Downsamples the data starting from `from_mag` until a magnification is `>= max(coarsest_mag)`.
        There are three different `sampling_modes`:
        - 'anisotropic' - The next magnification is chosen so that the width, height and depth of a downsampled voxel assimilate. For example, if the z resolution is worse than the x/y resolution, z won't be downsampled in the first downsampling step(s). As a basis for this method, the voxel_size from the datasource-properties.json is used.
        - 'isotropic' - Each dimension is downsampled equally.
        - 'constant_z' - The x and y dimensions are downsampled equally, but the z dimension remains the same.

        See `downsample_mag` for more information.

        Example:
        ```python
        from webknossos import SamplingModes

        # ...
        # let 'layer' be a `Layer` with only `Mag(1)`
        assert "1" in self.mags.keys()

        layer.downsample(
            coarsest_mag=Mag(4),
            sampling_mode=SamplingModes.ISOTROPIC
        )

        assert "2" in self.mags.keys()
        assert "4" in self.mags.keys()
        ```
        """
        if from_mag is None:
            assert (
                len(self.mags.keys()) > 0
            ), "Failed to downsample data because no existing mag was found."
            from_mag = max(self.mags.keys())

        assert (
            from_mag in self.mags.keys()
        ), f"Failed to downsample data. The from_mag ({from_mag.to_layer_name()}) does not exist."

        if coarsest_mag is None:
            coarsest_mag = calculate_default_coarsest_mag(self.bounding_box.size)

        sampling_mode = SamplingModes.parse(sampling_mode)

        if self._properties.bounding_box.size.z == 1:
            if sampling_mode != SamplingModes.CONSTANT_Z:
                warnings.warn(
                    "[INFO] The sampling_mode was changed to 'CONSTANT_Z'. Downsampling 2D data with a different sampling mode mixes in black and thus leads to darkened images."
                )
                sampling_mode = SamplingModes.CONSTANT_Z

        voxel_size: Optional[Tuple[float, float, float]] = None
        if sampling_mode == SamplingModes.ANISOTROPIC:
            voxel_size = self.dataset.voxel_size
        elif sampling_mode == SamplingModes.ISOTROPIC:
            voxel_size = None
        elif sampling_mode == SamplingModes.CONSTANT_Z:
            coarsest_mag_with_fixed_z = coarsest_mag.to_list()
            coarsest_mag_with_fixed_z[2] = from_mag.to_list()[2]
            coarsest_mag = Mag(coarsest_mag_with_fixed_z)
            voxel_size = None
        else:
            raise AttributeError(
                f"Downsampling failed: {sampling_mode} is not a valid SamplingMode ({SamplingModes.ANISOTROPIC}, {SamplingModes.ISOTROPIC}, {SamplingModes.CONSTANT_Z})"
            )

        dataset_to_align_with = self._get_dataset_from_align_with_other_layers(
            align_with_other_layers
        )
        mags_to_downsample = calculate_mags_to_downsample(
            from_mag, coarsest_mag, dataset_to_align_with, voxel_size
        )

        if len(set([max(m.to_list()) for m in mags_to_downsample])) != len(
            mags_to_downsample
        ):
            msg = (
                f"[INFO] The downsampling scheme contains multiple magnifications with the same maximum value. This is not supported by WEBKNOSSOS. "
                f"Consider using a different sampling mode (e.g. {SamplingModes.ISOTROPIC}). "
                f"The calculated downsampling scheme is: {[m.to_layer_name() for m in mags_to_downsample]}"
            )
            if force_sampling_scheme:
                warnings.warn(msg)
            else:
                raise RuntimeError(msg)

        for prev_mag, target_mag in zip(
            [from_mag] + mags_to_downsample[:-1], mags_to_downsample
        ):
            self.downsample_mag(
                from_mag=prev_mag,
                target_mag=target_mag,
                interpolation_mode=interpolation_mode,
                compress=compress,
                buffer_shape=buffer_shape,
                args=args,
                allow_overwrite=allow_overwrite,
                only_setup_mag=only_setup_mags,
                executor=executor,
            )

    def downsample_mag(
        self,
        from_mag: Mag,
        target_mag: Mag,
        interpolation_mode: str = "default",
        compress: bool = True,
        buffer_shape: Optional[Vec3Int] = None,
        args: Optional[Namespace] = None,  # deprecated
        allow_overwrite: bool = False,
        only_setup_mag: bool = False,
        executor: Optional[Executor] = None,
    ) -> None:
        """
        Performs a single downsampling step from `from_mag` to `target_mag`.

        The supported `interpolation_modes` are:
         - "median"
         - "mode"
         - "nearest"
         - "bilinear"
         - "bicubic"

        If allow_overwrite is True, an existing Mag may be overwritten.

        If only_setup_mag is True, the magnification is created, but left
        empty. This parameter can be used to prepare for parallel downsampling
        of multiple layers while avoiding parallel writes with outdated updates
        to the datasource-properties.json file.

        `executor` can be passed to allow distrubuted computation, parallelizing
        across chunks. `args` is deprecated.
        """
        self._dataset._ensure_writable()

        if args is not None:
            warn_deprecated(
                "args argument",
                "executor (e.g. via webknossos.utils.get_executor_for_args(args))",
            )

        assert (
            from_mag in self.mags.keys()
        ), f"Failed to downsample data. The from_mag ({from_mag.to_layer_name()}) does not exist."

        parsed_interpolation_mode = parse_interpolation_mode(
            interpolation_mode, self.category
        )

        assert from_mag <= target_mag
        assert (
            allow_overwrite or target_mag not in self.mags
        ), "The target mag already exists. Pass allow_overwrite=True if you want to overwrite it."

        prev_mag_view = self.mags[from_mag]

        mag_factors = target_mag.to_vec3_int() // from_mag.to_vec3_int()

        if target_mag in self.mags.keys() and allow_overwrite:
            target_mag_view = self.get_mag(target_mag)
        else:
            # initialize the new mag
            target_mag_view = self._initialize_mag_from_other_mag(
                target_mag, prev_mag_view, compress
            )

        if only_setup_mag:
            return

        bb_mag1 = self.bounding_box.align_with_mag(target_mag, ceil=True)

        # Get target view
        target_view = target_mag_view.get_view(
            absolute_offset=bb_mag1.topleft,
            size=bb_mag1.size,
        )

        source_view = prev_mag_view.get_view(
            absolute_offset=bb_mag1.topleft,
            size=bb_mag1.size,
            read_only=True,
        )

        # perform downsampling
        with get_executor_for_args(args, executor) as executor:
            if buffer_shape is None:
                buffer_shape = determine_buffer_shape(prev_mag_view.info)
            func = named_partial(
                downsample_cube_job,
                mag_factors=mag_factors,
                interpolation_mode=parsed_interpolation_mode,
                buffer_shape=buffer_shape,
            )

            source_view.for_zipped_chunks(
                # this view is restricted to the bounding box specified in the properties
                func,
                target_view=target_view,
                executor=executor,
                progress_desc=f"Downsampling layer {self.name} from Mag {from_mag} to Mag {target_mag}",
            )

    def redownsample(
        self,
        interpolation_mode: str = "default",
        compress: bool = True,
        buffer_shape: Optional[Vec3Int] = None,
        args: Optional[Namespace] = None,  # deprecated
        executor: Optional[Executor] = None,
    ) -> None:
        """
        Use this method to recompute downsampled magnifications after mutating data in the
        base magnification.
        """

        mags = sorted(self.mags.keys(), key=lambda m: m.to_list())
        if len(mags) <= 1:
            # No downsampled magnifications exist. Thus, there's nothing to do.
            return
        from_mag = mags[0]
        target_mags = mags[1:]
        self.downsample_mag_list(
            from_mag=from_mag,
            target_mags=target_mags,
            interpolation_mode=interpolation_mode,
            compress=compress,
            buffer_shape=buffer_shape,
            args=args,
            allow_overwrite=True,
            executor=executor,
        )

    def downsample_mag_list(
        self,
        from_mag: Mag,
        target_mags: List[Mag],
        interpolation_mode: str = "default",
        compress: bool = True,
        buffer_shape: Optional[Vec3Int] = None,
        args: Optional[Namespace] = None,  # deprecated
        allow_overwrite: bool = False,
        only_setup_mags: bool = False,
        executor: Optional[Executor] = None,
    ) -> None:
        """
        Downsamples the data starting at `from_mag` to each magnification in `target_mags` iteratively.

        See `downsample_mag` for more information.
        """
        assert (
            from_mag in self.mags.keys()
        ), f"Failed to downsample data. The from_mag ({from_mag}) does not exist."

        # The lambda function is important because 'sorted(target_mags)' would only sort by the maximum element per mag
        target_mags = sorted(target_mags, key=lambda m: m.to_list())

        for i in range(len(target_mags) - 1):
            assert np.less_equal(
                target_mags[i].to_np(), target_mags[i + 1].to_np()
            ).all(), (
                f"Downsampling failed: cannot downsample {target_mags[i].to_layer_name()} to {target_mags[i + 1].to_layer_name()}. "
                f"Check 'target_mags' ({', '.join([str(mag) for mag in target_mags])}): each pair of adjacent Mags results in a downsampling step."
            )

        source_mag = from_mag
        for target_mag in target_mags:
            self.downsample_mag(
                source_mag,
                target_mag,
                interpolation_mode=interpolation_mode,
                compress=compress,
                buffer_shape=buffer_shape,
                args=args,
                allow_overwrite=allow_overwrite,
                only_setup_mag=only_setup_mags,
                executor=executor,
            )
            source_mag = target_mag

    def upsample(
        self,
        from_mag: Mag,
        finest_mag: Mag = Mag(1),
        compress: bool = False,
        sampling_mode: Union[str, SamplingModes] = SamplingModes.ANISOTROPIC,
        align_with_other_layers: Union[bool, "Dataset"] = True,
        buffer_shape: Optional[Vec3Int] = None,
        buffer_edge_len: Optional[int] = None,
        args: Optional[Namespace] = None,  # deprecated
        executor: Optional[Executor] = None,
        *,
        min_mag: Optional[Mag] = None,
    ) -> None:
        """
        Upsamples the data starting from `from_mag` as long as the magnification is `>= finest_mag`.
        There are three different `sampling_modes`:
        - 'anisotropic' - The next magnification is chosen so that the width, height and depth of a downsampled voxel assimilate. For example, if the z resolution is worse than the x/y resolution, z won't be downsampled in the first downsampling step(s). As a basis for this method, the voxel_size from the datasource-properties.json is used.
        - 'isotropic' - Each dimension is downsampled equally.
        - 'constant_z' - The x and y dimensions are downsampled equally, but the z dimension remains the same.

        `min_mag` is deprecated, please use `finest_mag` instead.
        """
        self._dataset._ensure_writable()

        if args is not None:
            warn_deprecated(
                "args argument",
                "executor (e.g. via webknossos.utils.get_executor_for_args(args))",
            )

        assert (
            from_mag in self.mags.keys()
        ), f"Failed to upsample data. The from_mag ({from_mag.to_layer_name()}) does not exist."

        if min_mag is not None:
            warn_deprecated("upsample(min_mag=…)", "upsample(finest_mag=…)")
            assert finest_mag == Mag(
                1
            ), "Cannot set both min_mag and finest_mag, please only use finest_mag."
            finest_mag = min_mag

        sampling_mode = SamplingModes.parse(sampling_mode)

        voxel_size: Optional[Tuple[float, float, float]] = None
        if sampling_mode == SamplingModes.ANISOTROPIC:
            voxel_size = self.dataset.voxel_size
        elif sampling_mode == SamplingModes.ISOTROPIC:
            voxel_size = None
        elif sampling_mode == SamplingModes.CONSTANT_Z:
            finest_mag_with_fixed_z = finest_mag.to_list()
            finest_mag_with_fixed_z[2] = from_mag.to_list()[2]
            finest_mag = Mag(finest_mag_with_fixed_z)
            voxel_size = None
        else:
            raise AttributeError(
                f"Upsampling failed: {sampling_mode} is not a valid UpsamplingMode ({SamplingModes.ANISOTROPIC}, {SamplingModes.ISOTROPIC}, {SamplingModes.CONSTANT_Z})"
            )

        if buffer_shape is None and buffer_edge_len is not None:
            buffer_shape = Vec3Int.full(buffer_edge_len)

        dataset_to_align_with = self._get_dataset_from_align_with_other_layers(
            align_with_other_layers
        )
        mags_to_upsample = calculate_mags_to_upsample(
            from_mag, finest_mag, dataset_to_align_with, voxel_size
        )

        for prev_mag, target_mag in zip(
            [from_mag] + mags_to_upsample[:-1], mags_to_upsample
        ):
            assert prev_mag > target_mag
            assert target_mag not in self.mags

            prev_mag_view = self.mags[prev_mag]

            mag_factors = [
                t / s for (t, s) in zip(target_mag.to_list(), prev_mag.to_list())
            ]

            # initialize the new mag
            target_mag_view = self._initialize_mag_from_other_mag(
                target_mag, prev_mag_view, compress
            )

            # We need to make sure the layer's bounding box is aligned
            # with the previous mag. Otherwise, `for_zipped_chunks` will fail.
            # Saving the original layer bbox for later restore
            old_layer_bbox = self.bounding_box
            self.bounding_box = prev_mag_view.bounding_box
            # Get target view
            target_view = target_mag_view.get_view()

            # perform upsampling
            with get_executor_for_args(args, executor) as actual_executor:
                if buffer_shape is None:
                    buffer_shape = determine_buffer_shape(prev_mag_view.info)
                func = named_partial(
                    upsample_cube_job,
                    mag_factors=mag_factors,
                    buffer_shape=buffer_shape,
                )
                prev_mag_view.get_view().for_zipped_chunks(
                    # this view is restricted to the bounding box specified in the properties
                    func,
                    target_view=target_view,
                    executor=actual_executor,
                    progress_desc=f"Upsampling from Mag {prev_mag} to Mag {target_mag}",
                )
            # Restoring the original layer bbox
            self.bounding_box = old_layer_bbox

    def _setup_mag(self, mag: Mag) -> None:
        # This method is used to initialize the mag when opening the Dataset. This does not create e.g. the wk_header.

        mag_name = mag.to_layer_name()

        self._assert_mag_does_not_exist_yet(mag)

        try:
            cls_array = BaseArray.get_class(self._properties.data_format)
            info = cls_array.open(
                _find_mag_path_on_disk(self.dataset.path, self.name, mag_name)
            ).info
            self._mags[mag] = MagView(
                self,
                mag,
                info.chunk_shape,
                info.chunks_per_shard,
                info.compression_mode,
            )
            self._mags[mag]._read_only = self._dataset.read_only
        except ArrayException:
            logging.exception(
                f"Failed to setup magnification {mag_name}, which is specified in the datasource-properties.json:"
            )

    def _initialize_mag_from_other_mag(
        self, new_mag_name: Union[str, Mag], other_mag: MagView, compress: bool
    ) -> MagView:
        return self.add_mag(
            new_mag_name,
            chunk_shape=other_mag.info.chunk_shape,
            chunks_per_shard=other_mag.info.chunks_per_shard,
            compress=compress,
        )

    def __repr__(self) -> str:
        return f"Layer({repr(self.name)}, dtype_per_channel={self.dtype_per_channel}, num_channels={self.num_channels})"

    def _get_largest_segment_id_maybe(self) -> Optional[int]:
        return None

    @classmethod
    def _ensure_layer(cls, layer: Union[str, PathLike, "Layer"]) -> "Layer":
        if isinstance(layer, Layer):
            return layer
        else:
            # local import to prevent circular dependency
            from .dataset import Dataset

            layer_path = UPath(layer)
            return Dataset.open(layer_path.parent).get_layer(layer_path.name)


class SegmentationLayer(Layer):
    _properties: SegmentationLayerProperties

    @property
    def largest_segment_id(self) -> Optional[int]:
        return self._properties.largest_segment_id

    @largest_segment_id.setter
    def largest_segment_id(self, largest_segment_id: Optional[int]) -> None:
        self.dataset._ensure_writable()
        if largest_segment_id is not None and type(largest_segment_id) != int:
            assert largest_segment_id == int(
                largest_segment_id
            ), f"A non-integer value was passed for largest_segment_id ({largest_segment_id})."
            largest_segment_id = int(largest_segment_id)

        self._properties.largest_segment_id = largest_segment_id
        self.dataset._export_as_json()

    @property
    def category(self) -> LayerCategoryType:
        return SEGMENTATION_CATEGORY

    def _get_largest_segment_id_maybe(self) -> Optional[int]:
        return self.largest_segment_id

    def _get_largest_segment_id(self, view: View) -> int:
        return np.max(view.read(), initial=0)

    def refresh_largest_segment_id(
        self, chunk_shape: Optional[Vec3Int] = None, executor: Optional[Executor] = None
    ) -> None:
        """Sets the largest segment id to the highest value in the data.
        largest_segment_id is set to `None` if the data is empty."""

        try:
            chunk_results = self.get_finest_mag().map_chunk(
                self._get_largest_segment_id,
                chunk_shape=chunk_shape,
                executor=executor,
                progress_desc="Searching largest segment id",
            )
            self.largest_segment_id = max(chunk_results)
        except ValueError:
            self.largest_segment_id = None



"""This module converts a NIFTI image or a folder with NIFTI images to a WEBKNOSSOS dataset."""

import logging
from pathlib import Path
from typing import Any, Optional, Tuple, Union, cast

import nibabel as nib
import numpy as np
import typer
from sklearn.preprocessing import LabelEncoder
from typing_extensions import Annotated

from ..dataset import DataFormat, Dataset, LayerCategoryType
from ..dataset.defaults import DEFAULT_CHUNK_SHAPE, DEFAULT_CHUNKS_PER_SHARD
from ..geometry import BoundingBox, Vec3Int
from ..utils import time_start, time_stop
from ._utils import (
    Vec2Int,
    VoxelSize,
    pad_or_crop_to_size_and_topleft,
    parse_bbox,
    parse_path,
    parse_vec3int,
    parse_voxel_size,
)


def to_target_datatype(
    data: np.ndarray,
    target_dtype: Union[type, str, np.dtype],
    is_segmentation_layer: bool,
) -> np.ndarray:
    """Cast data into target datatype."""

    if is_segmentation_layer:
        original_shape = data.shape
        label_encoder = LabelEncoder()
        return (
            label_encoder.fit_transform(data.ravel())
            .reshape(original_shape)
            .astype(np.dtype(target_dtype))
        )

    factor: Any
    if data.dtype == np.dtype("float32"):
        factor = data.max()
    elif data.dtype == np.dtype("float64"):
        factor = data.max() / np.iinfo(target_dtype).max
    else:
        factor = np.iinfo(data.dtype).max / np.iinfo(target_dtype).max

    if data.max() == 0:
        logging.warning("Not rescaling data since maximum is 0")
        factor = 1

    return (data / factor).astype(np.dtype(target_dtype))


def convert_nifti(
    source_nifti_path: Path,
    target_path: Path,
    layer_name: str,
    dtype: str,
    voxel_size: Optional[VoxelSize],
    data_format: DataFormat,
    chunk_shape: Vec3Int,
    chunks_per_shard: Vec3Int,
    is_segmentation_layer: bool = False,
    bbox_to_enforce: Optional[BoundingBox] = None,
    use_orientation_header: bool = False,
    flip_axes: Optional[Union[int, Tuple[int, ...]]] = None,
) -> None:
    """Converts a single NIFTI file into a WEBKNOSSOS dataset."""

    shard_shape = chunk_shape * chunks_per_shard
    time_start(f"Converting of {source_nifti_path}")

    source_nifti = nib.load(str(source_nifti_path.resolve()))

    if use_orientation_header:
        # Get canonical representation of data to incorporate
        # encoded transformations. Needs to be flipped later
        # to match the coordinate system of WKW.
        source_nifti = nib.funcs.as_closest_canonical(source_nifti, enforce_diag=False)

    cube_data = np.array(source_nifti.get_fdata())  # type:ignore

    category_type: LayerCategoryType = (
        "segmentation" if is_segmentation_layer else "color"
    )
    logging.debug("Assuming %s as layer type for %s", category_type, layer_name)

    if len(source_nifti.shape) == 3:  # type:ignore
        cube_data = cube_data.reshape((1,) + source_nifti.shape)  # type:ignore

    elif len(source_nifti.shape) == 4:  # type:ignore
        cube_data = np.transpose(cube_data, (3, 0, 1, 2))

    else:
        logging.warning(
            "Converting of %s failed! Too many or too less dimensions",
            source_nifti_path,
        )

        return

    if use_orientation_header:
        # Flip y and z to transform data into wkw's coordinate system.
        cube_data = np.flip(cube_data, (2, 3))

    if flip_axes:
        cube_data = np.flip(cube_data, flip_axes)

    if voxel_size is None:
        voxel_size = tuple(map(float, source_nifti.header["pixdim"][:3]))  # type:ignore

    logging.info("Using voxel_size: %s", voxel_size)
    cube_data = to_target_datatype(cube_data, dtype, is_segmentation_layer)

    # everything needs to be padded to
    if bbox_to_enforce is not None:
        target_topleft = np.array((0,) + tuple(bbox_to_enforce.topleft))
        target_size = np.array((1,) + tuple(bbox_to_enforce.size))

        cube_data = pad_or_crop_to_size_and_topleft(
            cube_data, target_size, target_topleft
        )

    # Writing wkw compressed requires files of shape (shard_shape, shard_shape, shard_shape)
    # Pad data accordingly
    padding_offset = shard_shape - np.array(cube_data.shape[1:4]) % shard_shape
    cube_data = np.pad(
        cube_data,
        (
            (0, 0),
            (0, int(padding_offset[0])),
            (0, int(padding_offset[1])),
            (0, int(padding_offset[2])),
        ),
    )

    wk_ds = Dataset(
        target_path,
        voxel_size=cast(Tuple[float, float, float], voxel_size or (1, 1, 1)),
        exist_ok=True,
    )
    wk_layer = (
        wk_ds.get_or_add_layer(
            layer_name,
            category_type,
            dtype_per_layer=np.dtype(dtype),
            data_format=data_format,
            largest_segment_id=int(np.max(cube_data) + 1),
        )
        if is_segmentation_layer
        else wk_ds.get_or_add_layer(
            layer_name,
            category_type,
            data_format=data_format,
            dtype_per_layer=np.dtype(dtype),
        )
    )
    wk_mag = wk_layer.get_or_add_mag(
        "1", chunk_shape=chunk_shape, chunks_per_shard=chunks_per_shard
    )
    wk_mag.write(cube_data)

    time_stop(f"Converting of {source_nifti_path}")


def convert_folder_nifti(
    source_folder_path: Path,
    target_path: Path,
    color_subpath: Optional[str],
    segmentation_subpath: Optional[str],
    voxel_size: Optional[VoxelSize],
    data_format: DataFormat,
    chunk_shape: Vec3Int,
    chunks_per_shard: Vec3Int,
    use_orientation_header: bool = False,
    bbox_to_enforce: Optional[BoundingBox] = None,
    flip_axes: Optional[Union[int, Tuple[int, ...]]] = None,
) -> None:
    """Converts a folder of NIFTI files into WEBKNOSSOS dataset."""
    paths = list(source_folder_path.rglob("**/*.nii"))

    color_path = None
    segmentation_path = None
    if color_subpath is not None:
        color_path = target_path / color_subpath
        if color_path not in paths:
            logging.warning(
                "Specified color file %s not in source path %s!",
                color_path,
                source_folder_path,
            )

    if segmentation_subpath is not None:
        segmentation_path = target_path / segmentation_subpath
        if segmentation_path not in paths:
            logging.warning(
                "Specified segmentation_file file %s not in source path %s!",
                segmentation_path,
                source_folder_path,
            )

    logging.info("Segmentation file will also use uint8 as a datatype.")

    for path in paths:
        if path == color_path:
            convert_nifti(
                path,
                target_path,
                "color",
                "uint8",
                voxel_size,
                data_format,
                chunk_shape,
                chunks_per_shard,
                is_segmentation_layer=False,
                bbox_to_enforce=bbox_to_enforce,
                use_orientation_header=use_orientation_header,
                flip_axes=flip_axes,
            )
        elif path == segmentation_path:
            convert_nifti(
                path,
                target_path,
                "segmentation",
                "uint8",
                voxel_size,
                data_format,
                chunk_shape,
                chunks_per_shard,
                is_segmentation_layer=True,
                bbox_to_enforce=bbox_to_enforce,
                use_orientation_header=use_orientation_header,
                flip_axes=flip_axes,
            )
        else:
            convert_nifti(
                path,
                target_path,
                path.stem,
                "uint8",
                voxel_size,
                data_format,
                chunk_shape,
                chunks_per_shard,
                is_segmentation_layer=False,
                bbox_to_enforce=bbox_to_enforce,
                use_orientation_header=use_orientation_header,
                flip_axes=flip_axes,
            )


def main(
    *,
    source: Annotated[
        Any,
        typer.Argument(
            help="Path to your image data.",
            show_default=False,
            parser=parse_path,
        ),
    ],
    target: Annotated[
        Any,
        typer.Argument(
            help="Target path to save your WEBKNOSSOS dataset.",
            show_default=False,
            parser=parse_path,
        ),
    ],
    layer_name: Annotated[
        str,
        typer.Option(help="Name of the cubed layer (color or segmentation)"),
    ] = "color",
    voxel_size: Annotated[
        Optional[VoxelSize],
        typer.Option(
            help="The size of one voxel in source data in nanometers. \
Should be a comma seperated string (e.g. 11.0,11.0,20.0).",
            parser=parse_voxel_size,
            metavar="VOXEL_SIZE",
        ),
    ] = None,
    dtype: Annotated[
        str, typer.Option(help="Target datatype (e.g. uint8, uint16, uint32)")
    ] = "uint8",
    data_format: Annotated[
        DataFormat,
        typer.Option(
            help="Data format to store the target dataset in.",
        ),
    ] = "wkw",  # type:ignore
    chunk_shape: Annotated[
        Vec3Int,
        typer.Option(
            help="Number of voxels to be stored as a chunk in the output format "
            "(e.g. `32` or `32,32,32`).",
            parser=parse_vec3int,
            metavar="Vec3Int",
        ),
    ] = DEFAULT_CHUNK_SHAPE,
    chunks_per_shard: Annotated[
        Vec3Int,
        typer.Option(
            help="Number of chunks to be stored as a shard in the output format "
            "(e.g. `32` or `32,32,32`).",
            parser=parse_vec3int,
            metavar="Vec3Int",
        ),
    ] = DEFAULT_CHUNKS_PER_SHARD,
    enforce_bounding_box: Annotated[
        Optional[BoundingBox],
        typer.Option(
            help="The BoundingBox to which the input data should be written. "
            "If the input data is too small, it will be padded. If it's too large, "
            "it will be cropped. The input format is x,y,z,width,height,depth.",
            parser=parse_bbox,
            metavar="BBOX",
        ),
    ] = None,
    color_file: Annotated[
        Optional[str],
        typer.Option(
            help="When converting folder, name of file to become color layer."
        ),
    ] = None,
    segmentation_file: Annotated[
        Optional[str],
        typer.Option(
            help="When converting folder, name of file to become segmentation layer."
        ),
    ] = None,
    use_orientation_header: Annotated[
        bool,
        typer.Option(
            help="Use orientation information from header to interpret the input data \
(should be tried if output orientation seems to be wrong)."
        ),
    ] = False,
    is_segmentation_layer: Annotated[
        bool,
        typer.Option(
            help="When converting one layer, signals whether layer is segmentation layer. \
When converting a folder, this option is ignored."
        ),
    ] = False,
    flip_axes: Annotated[
        Optional[Vec2Int],
        typer.Option(
            help="The axes at which should be flipped. \
Input format is a comma separated list of axis indices. \
For example, 1,2,3 will flip the x, y and z axes.",
            parser=parse_vec3int,
            metavar="Vec3Int",
        ),
    ] = None,
) -> None:
    """Converts a NIFTI file or a folder of NIFTI files into a WEBKNOSSOS dataset."""

    if flip_axes is not None:
        for index in flip_axes:
            assert (
                0 <= index <= 3
            ), "flip_axes parameter must only contain indices between 0 and 3."

    if source.is_dir():
        convert_folder_nifti(
            source,
            target,
            color_file,
            segmentation_file,
            voxel_size=voxel_size,
            data_format=data_format,
            chunk_shape=chunk_shape,
            chunks_per_shard=chunks_per_shard,
            bbox_to_enforce=enforce_bounding_box,
            use_orientation_header=use_orientation_header,
            flip_axes=flip_axes,
        )
    else:
        convert_nifti(
            source,
            target,
            layer_name,
            dtype,
            voxel_size=voxel_size,
            data_format=data_format,
            chunk_shape=chunk_shape,
            chunks_per_shard=chunks_per_shard,
            is_segmentation_layer=is_segmentation_layer,
            bbox_to_enforce=enforce_bounding_box,
            use_orientation_header=use_orientation_header,
            flip_axes=flip_axes,
        )



"""
# Dataset API

The high-level dataset API automatically reads and writes meta information for any dataset and updates them if necessary, such as the `datasource-properties.json`.

A dataset (`webknossos.dataset.dataset.Dataset`) is the entry-point for this API.
The dataset stores the data on disk in `.wkw`-files (see [WEBKNOSSOS-wrap (wkw)](https://github.com/scalableminds/webknossos-wrap)).

Each dataset consists of one or more layers (webknossos.dataset.layer.Layer), which themselves can comprise multiple magnifications (webknossos.dataset.mag_view.MagView).
"""

from .data_format import DataFormat
from .dataset import Dataset, RemoteDataset
from .layer import Layer, SegmentationLayer
from .layer_categories import COLOR_CATEGORY, SEGMENTATION_CATEGORY, LayerCategoryType
from .mag_view import MagView
from .remote_folder import RemoteFolder
from .sampling_modes import SamplingModes
from .view import View



from pathlib import Path
from typing import Any, Callable, Dict, List, Sequence

import attr


class CallbackDispatcher:
    """Dispatch callbacks to registered targets."""

    def __init__(self) -> None:
        self.targets: List[Callable] = []

    def register(self, callback: Callable) -> None:
        self.targets.append(callback)

    def trigger(self, *args: Any, **kwargs: Any) -> None:
        """Trigger this dispatcher.

        All arguments are passed through to the registered callbacks.
        """
        for callback in self.targets:
            callback(*args, **kwargs)


@attr.frozen
class Config:
    """The configuration for a resumable session."""

    target: str
    chunk_size: int
    simultaneous_uploads: int
    headers: Dict[str, Any]
    test_chunks: bool
    max_chunk_retries: int
    permanent_errors: Sequence[int]
    additional_query_params: Dict[str, Any]
    generate_unique_identifier: Callable[[Path, Path], str]



import json
from typing import Dict, List, Optional, Tuple

import httpx

from webknossos.client.api_client.models import (
    ApiAnnotation,
    ApiAnnotationUploadResult,
    ApiDataset,
    ApiDataStore,
    ApiDataStoreToken,
    ApiFolderWithParent,
    ApiLoggedTimeGroupedByMonth,
    ApiNmlTaskParameters,
    ApiProject,
    ApiSharingToken,
    ApiShortLink,
    ApiTask,
    ApiTaskCreationResult,
    ApiTaskParameters,
    ApiTeam,
    ApiUser,
    ApiWkBuildInfo,
)

from ...utils import time_since_epoch_in_ms
from ._abstract_api_client import AbstractApiClient


class WkApiClient(AbstractApiClient):
    # Client to use the HTTP API of WEBKNOSSOS servers.
    # When adding a method here, use the utility methods from AbstractApiClient
    # and add more as needed.
    # Methods here are prefixed with the domain, e.g. dataset_update_teams (not update_dataset_teams)

    def __init__(
        self,
        base_wk_url: str,
        timeout_seconds: float,
        headers: Optional[Dict[str, str]] = None,
        webknossos_api_version: int = 5,
    ):
        super().__init__(timeout_seconds, headers)
        self.webknossos_api_version = webknossos_api_version
        self.base_wk_url = base_wk_url

    @property
    def url_prefix(self) -> str:
        return f"{self.base_wk_url}/api/v{self.webknossos_api_version}"

    def health(self) -> None:
        route = "/health"
        self._get(route)

    def build_info(self) -> ApiWkBuildInfo:
        route = "/buildinfo"
        return self._get_json(route, ApiWkBuildInfo)

    def short_link_by_key(self, key: str) -> ApiShortLink:
        route = f"/shortLinks/byKey/{key}"
        return self._get_json(route, ApiShortLink)

    def dataset_info(
        self,
        organization_name: str,
        dataset_name: str,
        sharing_token: Optional[str] = None,
    ) -> ApiDataset:
        route = f"/datasets/{organization_name}/{dataset_name}"
        return self._get_json(route, ApiDataset, query={"sharingToken": sharing_token})

    def dataset_list(
        self, is_active: Optional[bool], organization_name: Optional[str]
    ) -> List[ApiDataset]:
        route = "/datasets"
        return self._get_json(
            route,
            List[ApiDataset],
            query={"isActive": is_active, "organizationName": organization_name},
        )

    def dataset_update_teams(
        self, organization_name: str, dataset_name: str, team_ids: List[str]
    ) -> None:
        route = f"/datasets/{organization_name}/{dataset_name}/teams"
        self._patch_json(route, team_ids)

    def dataset_update(
        self, organization_name: str, dataset_name: str, updated_dataset: ApiDataset
    ) -> None:
        route = f"/datasets/{organization_name}/{dataset_name}"
        self._patch_json(route, updated_dataset)

    def dataset_sharing_token(
        self, organization_name: str, dataset_name: str
    ) -> ApiSharingToken:
        route = f"/datasets/{organization_name}/{dataset_name}/sharingToken"
        return self._get_json(route, ApiSharingToken)

    def dataset_assert_new_name_is_valid(
        self, organization_name: str, dataset_name: str
    ) -> None:
        route = f"/datasets/{organization_name}/{dataset_name}/isValidNewName"
        self._get(route)

    def datastore_list(self) -> List[ApiDataStore]:
        route = "/datastores"
        return self._get_json(route, List[ApiDataStore])

    def project_info_by_name(self, project_name: str) -> ApiProject:
        route = f"/projects/byName/{project_name}"
        return self._get_json(route, ApiProject)

    def project_info_by_id(self, project_id: str) -> ApiProject:
        route = f"/projects/{project_id}"
        return self._get_json(route, ApiProject)

    def task_infos_by_project_id_paginated(
        self, project_id: str, limit: Optional[int], page_number: Optional[int]
    ) -> Tuple[List[ApiTask], int]:
        route = f"/projects/{project_id}/tasks"
        return self._get_json_paginated(route, List[ApiTask], limit, page_number)

    def annotation_info(self, annotation_id: str) -> ApiAnnotation:
        route = f"/annotations/{annotation_id}/info"
        return self._get_json(
            route, ApiAnnotation, query={"timestamp": time_since_epoch_in_ms()}
        )

    def annotation_download(
        self, annotation_id: str, skip_volume_data: bool
    ) -> Tuple[bytes, str]:
        route = f"/annotations/{annotation_id}/download"
        return self._get_file(route, query={"skipVolumeData": skip_volume_data})

    def annotation_upload(
        self, file_body: bytes, filename: str, createGroupForEachFile: bool
    ) -> ApiAnnotationUploadResult:
        route = "/annotations/upload"
        data: httpx._types.RequestData = {
            "createGroupForEachFile": createGroupForEachFile
        }
        files: httpx._types.RequestFiles = {
            filename: (filename, file_body),
        }
        return self.post_multipart_with_json_response(
            route, ApiAnnotationUploadResult, data, files
        )

    def annotation_infos_by_task(self, task_id: str) -> List[ApiAnnotation]:
        route = f"/tasks/{task_id}/annotations"
        return self._get_json(route, List[ApiAnnotation])

    def task_info(self, task_id: str) -> ApiTask:
        route = f"/tasks/{task_id}"
        return self._get_json(route, ApiTask)

    def folder_tree(self) -> List[ApiFolderWithParent]:
        route = "/folders/tree"
        return self._get_json(route, List[ApiFolderWithParent])

    def user_by_id(self, user_id: str) -> ApiUser:
        route = f"/users/{user_id}"
        return self._get_json(route, ApiUser)

    def user_current(self) -> ApiUser:
        route = "/user"
        return self._get_json(route, ApiUser)

    def user_list(self) -> List[ApiUser]:
        route = "/users"
        return self._get_json(route, List[ApiUser])

    def user_logged_time(self, user_id: str) -> ApiLoggedTimeGroupedByMonth:
        route = f"/users/{user_id}/loggedTime"
        return self._get_json(route, ApiLoggedTimeGroupedByMonth)

    def team_list(self) -> List[ApiTeam]:
        route = "/teams"
        return self._get_json(route, List[ApiTeam])

    def token_generate_for_data_store(self) -> ApiDataStoreToken:
        route = "/userToken/generate"
        return self._post_with_json_response(route, ApiDataStoreToken)

    def tasks_create(
        self, task_parameters: List[ApiTaskParameters]
    ) -> ApiTaskCreationResult:
        route = "/tasks"
        return self._post_json_with_json_response(
            route, task_parameters, ApiTaskCreationResult
        )

    def tasks_create_from_files(
        self,
        nml_task_parameters: ApiNmlTaskParameters,
        annotation_files: List[Tuple[str, bytes]],
    ) -> ApiTaskCreationResult:
        route = "/tasks/createFromFiles"
        data: httpx._types.RequestData = {
            "formJSON": json.dumps(self._prepare_for_json(nml_task_parameters))
        }
        files: httpx._types.RequestFiles = {
            filename: (filename, file_body) for filename, file_body in annotation_files
        }
        return self.post_multipart_with_json_response(
            route, ApiTaskCreationResult, multipart_data=data, files=files
        )



import copy
from typing import TYPE_CHECKING, Iterator, Optional, Set, Tuple, Union, cast

import attr
from boltons.strutils import unit_len

import webknossos._nml as wknml

from ..utils import warn_deprecated
from .tree import Tree

if TYPE_CHECKING:
    from .node import Node
    from .skeleton import Skeleton


Vector3 = Tuple[float, float, float]
Vector4 = Tuple[float, float, float, float]
GroupOrTree = Union["Group", Tree]


@attr.define()
class Group:
    _id: int = attr.ib(init=False)
    name: str
    _child_groups: Set["Group"] = attr.ib(
        factory=set,
        init=False,
        repr=lambda children: f"<{unit_len(children, 'child group')}>",
    )
    _child_trees: Set[Tree] = attr.ib(
        factory=set,
        init=False,
        repr=lambda children: f"<{unit_len(children, 'child tree')}>",
    )
    _skeleton: "Skeleton" = attr.ib(eq=False, repr=False)
    _enforced_id: Optional[int] = attr.ib(None, eq=False, repr=False)

    @classmethod
    def _set_init_docstring(cls) -> None:
        Group.__init__.__doc__ = """
        To create a group, it is recommended to use `webknossos.skeleton.skeleton.Skeleton.add_group` or
        `Group.add_group`. That way, the newly created group
        is automatically attached as a child to the object the method was
        called on.

        A small usage example:

        ```python
        subgroup = group.add_group("a subgroup")
        tree = subgroup.add_tree("a tree")
        ```
        """

    def __attrs_post_init__(self) -> None:
        if self._enforced_id is not None:
            self._id = self._enforced_id
        else:
            self._id = self._skeleton._element_id_generator.__next__()

    @property
    def id(self) -> int:
        """Read-only property."""
        return self._id

    def add_tree(
        self,
        name_or_tree: Union[str, Tree],
        color: Optional[Union[Vector4, Vector3]] = None,
        _enforced_id: Optional[int] = None,
    ) -> Tree:
        """Adds a tree to the current group. If the first parameter is a string,
        a new tree will be added with the provided name and color if specified.
        Otherwise, the first parameter is assumed to be a tree object (e.g., from
        another skeleton). A copy of that tree will then be added. If the id
        of the tree already exists, a new id will be generated."""

        if color is not None and len(color) == 3:
            color = cast(Optional[Vector4], color + (1.0,))
        color = cast(Optional[Vector4], color)

        if type(name_or_tree) is str:
            name = name_or_tree
            new_tree = Tree(
                name=name,
                color=color,
                group=self,
                skeleton=self._skeleton,
                enforced_id=_enforced_id,
            )
            self._child_trees.add(new_tree)

            return new_tree
        else:
            tree = cast(Tree, name_or_tree)
            new_tree = copy.deepcopy(tree)

            if color is not None:
                new_tree.color = color

            if _enforced_id is not None:
                assert not self.has_tree_id(
                    _enforced_id
                ), "A tree with the specified _enforced_id already exists in this group."
                new_tree._id = _enforced_id

            if self.has_tree_id(tree.id):
                new_tree._id = self._skeleton._element_id_generator.__next__()

            new_tree.group = self
            new_tree.skeleton = self._skeleton

            self._child_trees.add(new_tree)
            return new_tree

    def add_graph(
        self,
        name: str,
        color: Optional[Union[Vector4, Vector3]] = None,
        _enforced_id: Optional[int] = None,
    ) -> Tree:
        """Deprecated, please use `add_tree`."""
        warn_deprecated("add_graph()", "add_tree()")
        return self.add_tree(name_or_tree=name, color=color, _enforced_id=_enforced_id)

    def remove_tree_by_id(self, tree_id: int) -> None:
        self._child_trees.remove(self.get_tree_by_id(tree_id))

    @property
    def children(self) -> Iterator[GroupOrTree]:
        """Returns all (immediate) children (groups and trees) as an iterator."""
        yield from self.trees
        yield from self.groups

    @property
    def trees(self) -> Iterator[Tree]:
        """Returns all (immediate) tree children as an iterator.
        Use flattened_trees if you need also need trees within subgroups."""
        return (child for child in self._child_trees)

    @property
    def graphs(self) -> Iterator[Tree]:
        """Deprecated, please use `trees`."""
        warn_deprecated("graphs", "trees")
        return self.trees

    @property
    def groups(self) -> Iterator["Group"]:
        """Returns all (immediate) group children as an iterator.
        Use flattened_groups if you need also need groups within subgroups."""
        return (child for child in self._child_groups)

    def add_group(
        self,
        name: str,
        _enforced_id: Optional[int] = None,
    ) -> "Group":
        """Adds a (sub) group to the current group with the provided name."""
        new_group = Group(name, skeleton=self._skeleton, enforced_id=_enforced_id)
        self._child_groups.add(new_group)
        return new_group

    def get_total_node_count(self) -> int:
        """Returns the total number of nodes of all trees within this group (and its subgroups)."""
        return sum(tree.number_of_nodes() for tree in self.flattened_trees())

    def get_max_tree_id(self) -> int:
        """Returns the highest tree id of all trees within this group (and its subgroups)."""
        return max((tree.id for tree in self.flattened_trees()), default=0)

    def get_max_graph_id(self) -> int:
        """Deprecated, please use `get_max_tree_id`."""
        warn_deprecated("get_max_graph_id()", "get_max_tree_id()")
        return self.get_max_tree_id()

    def get_max_node_id(self) -> int:
        """Returns the highest node id of all nodes of all trees within this group (and its subgroups)."""
        return max(
            (tree.get_max_node_id() for tree in self.flattened_trees()),
            default=0,
        )

    def flattened_trees(self) -> Iterator[Tree]:
        """Returns an iterator of all trees within this group (and its subgroups)."""
        yield from self.trees
        for group in self.groups:
            yield from group.flattened_trees()

    def flattened_graphs(self) -> Iterator[Tree]:
        """Deprecated, please use `flattened_trees`."""
        warn_deprecated("flattened_graphs()", "flattened_trees()")
        return self.flattened_trees()

    def flattened_groups(self) -> Iterator["Group"]:
        """Returns an iterator of all groups within this group (and its subgroups)."""
        for group in self.groups:
            yield group
            yield from group.flattened_groups()

    def get_node_by_id(self, node_id: int) -> "Node":
        """Returns the node which has the specified node id."""
        for tree in self.flattened_trees():
            if tree.has_node(node_id):
                return tree.get_node_by_id(node_id)

        raise ValueError("Node id not found")

    def get_tree_by_id(self, tree_id: int) -> Tree:
        """Returns the tree which has the specified tree id."""
        # Todo: Use hashed access if it turns out to be worth it? pylint: disable=fixme
        for tree in self.flattened_trees():
            if tree.id == tree_id:
                return tree
        raise ValueError(f"No tree with id {tree_id} was found")

    def has_tree_id(self, tree_id: int) -> bool:
        """Returns true if this group (or a subgroup) contains a tree with the given id."""
        try:
            self.get_tree_by_id(tree_id)
            return True
        except ValueError:
            return False

    def get_graph_by_id(self, graph_id: int) -> Tree:
        """Deprecated, please use `get_tree_by_id`."""
        warn_deprecated("get_graph_by_id()", "get_tree_by_id()")
        return self.get_tree_by_id(graph_id)

    def get_group_by_id(self, group_id: int) -> "Group":
        """Returns the group which has the specified group id."""
        # Todo: Use hashed access if it turns out to be worth it? pylint: disable=fixme
        for group in self.flattened_groups():
            if group.id == group_id:
                return group
        raise ValueError(f"No group with id {group_id} was found")

    def as_nml_group(self) -> wknml.Group:
        """Returns a named tuple representation of this group."""
        return wknml.Group(
            self.id,
            self.name,
            children=[g.as_nml_group() for g in self._child_groups],
        )

    def __hash__(self) -> int:
        return self._id


Group._set_init_docstring()



import warnings
from logging import info
from os import getpid
from types import TracebackType
from typing import TYPE_CHECKING, Generator, Optional, Type

import numpy as np

if TYPE_CHECKING:
    from ..view import View

from ...geometry import BoundingBox, Vec3IntLike
from ...utils import get_chunks


class BufferedSliceReader:
    def __init__(
        self,
        view: "View",
        offset: Optional[Vec3IntLike] = None,
        size: Optional[Vec3IntLike] = None,
        # buffer_size specifies, how many slices should be aggregated until they are flushed.
        buffer_size: int = 32,
        dimension: int = 2,  # z
        *,
        relative_bounding_box: Optional[BoundingBox] = None,  # in mag1
        absolute_bounding_box: Optional[BoundingBox] = None,  # in mag1
        use_logging: bool = False,
    ) -> None:
        """see `View.get_buffered_slice_reader()`"""

        self.view = view
        self.buffer_size = buffer_size
        self.dtype = self.view.get_dtype()
        assert 0 <= dimension <= 2
        self.dimension = dimension
        self.use_logging = use_logging
        if offset is not None and size is not None:
            warnings.warn(
                "[DEPRECATION] Using offset and size for a buffered slice reader is deprecated. "
                + "Please use the parameter relative_bounding_box or absolute_bounding_box in Mag(1) instead.",
                DeprecationWarning,
            )
            assert relative_bounding_box is None and absolute_bounding_box is None
            absolute_bounding_box = BoundingBox(offset, size).from_mag_to_mag1(view.mag)
            offset = None
            size = None

        assert (
            offset is None and size is None
        ), "You have to set both offset and size or none of both."
        if relative_bounding_box is None and absolute_bounding_box is None:
            absolute_bounding_box = view.bounding_box
        if relative_bounding_box is not None:
            assert absolute_bounding_box is None
            absolute_bounding_box = relative_bounding_box.offset(
                view.bounding_box.topleft
            )

        assert absolute_bounding_box is not None
        self.bbox_current_mag = absolute_bounding_box.in_mag(view.mag)

    def _get_slice_generator(self) -> Generator[np.ndarray, None, None]:
        for batch in get_chunks(
            list(
                range(
                    self.bbox_current_mag.topleft[self.dimension],
                    self.bbox_current_mag.bottomright[self.dimension],
                )
            ),
            self.buffer_size,
        ):
            n_slices = len(batch)
            batch_start_idx = batch[0]

            assert (
                n_slices <= self.buffer_size
            ), f"n_slices should at most be batch_size, but {n_slices} > {self.buffer_size}"

            bbox_offset = self.bbox_current_mag.topleft
            bbox_size = self.bbox_current_mag.size

            buffer_bounding_box = BoundingBox.from_tuple2(
                (
                    bbox_offset[: self.dimension]
                    + (batch_start_idx,)
                    + bbox_offset[self.dimension + 1 :],
                    bbox_size[: self.dimension]
                    + (n_slices,)
                    + bbox_size[self.dimension + 1 :],
                )
            )

            if self.use_logging:
                info(
                    f"({getpid()}) Reading {n_slices} slices at position {batch_start_idx}."
                )
            data = self.view.read(
                absolute_bounding_box=buffer_bounding_box.from_mag_to_mag1(
                    self.view.mag
                )
            )

            for current_slice in np.rollaxis(
                data, self.dimension + 1
            ):  # The '+1' is important because the first dimension is the channel
                yield current_slice

    def __enter__(self) -> Generator[np.ndarray, None, None]:
        return self._get_slice_generator()

    def __exit__(
        self,
        _type: Optional[Type[BaseException]],
        _value: Optional[BaseException],
        _tb: Optional[TracebackType],
    ) -> None:
        ...



from .branchpoint import Branchpoint
from .comment import Comment
from .edge import Edge
from .group import Group
from .meta import Meta
from .nml import Nml
from .node import Node
from .parameters import Parameters
from .segment import Segment
from .tree import Tree
from .volume import Volume



from typing import TYPE_CHECKING, Optional, Tuple

import attr

from ..geometry import Vec3Int

if TYPE_CHECKING:
    from .skeleton import Skeleton

Vec3Float = Tuple[float, float, float]


# Defining an order on nodes is necessary to allow to sort them,
# which is used in the tree's equality check, see Tree.__eq__().
@attr.define(order=True)
class Node:
    position: Vec3Int
    _skeleton: "Skeleton" = attr.ib(eq=False, repr=False, order=False)
    _id: int = attr.ib(init=False)
    comment: Optional[str] = None
    radius: Optional[float] = None
    rotation: Optional[Vec3Float] = None
    inVp: Optional[int] = None
    inMag: Optional[int] = None
    bitDepth: Optional[int] = None
    interpolation: Optional[bool] = None
    time: Optional[int] = None

    is_branchpoint: bool = False
    branchpoint_time: Optional[int] = None
    _enforced_id: Optional[int] = attr.ib(None, eq=False, repr=False)

    @classmethod
    def _set_init_docstring(cls) -> None:
        Node.__init__.__doc__ = """
        To create a node, it is recommended to use `Tree.add_node`. That way,
        the newly created group is automatically attached as a child to the
        tree.

        A small usage example:

        ```python
        tree = skeleton.add_tree("a tree")
        node_1 = tree.add_node(position=(0, 0, 0), comment="node 1")
        node_2 = tree.add_node(position=(100, 100, 100), comment="node 2")

        tree.add_edge(node_1, node_2)
        ```
        """

    def __attrs_post_init__(self) -> None:
        if self._enforced_id is not None:
            self._id = self._enforced_id
        else:
            self._id = self._skeleton._element_id_generator.__next__()

    @property
    def id(self) -> int:
        return self._id

    def __hash__(self) -> int:
        return self._id


Node._set_init_docstring()



import uuid
from concurrent.futures import Future, ThreadPoolExecutor, as_completed
from functools import partial
from os import PathLike
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, Sequence, Union

import httpx

from .chunk import resolve_chunk
from .file import ResumableFile
from .util import CallbackDispatcher, Config

MiB = 1024 * 1024


class Resumable:
    """A resumable-httpx upload client.

    Parameters
    ----------
    target : str
        The URL of the resumable upload target
    chunk_size : int, optional
        The size, in bytes, of file chunks to be uploaded
    simultaneous_uploads : int, optional
        The number of file chunk uploads to attempt at once
    headers : dict, optional
        A dictionary of additional HTTP headers to include in requests
    test_chunks : bool
        Flag indicating if the client should check with the server if a chunk
        already exists with a GET request prior to attempting to upload the
        chunk with a POST
    max_chunk_retries : int, optional
        The number of times to retry uploading a chunk
    permanent_errors : collection of int, optional
        HTTP status codes that indicate the upload of a chunk has failed and
        should not be retried

    Attributes
    ----------
    file_added : resumable.util.CallbackDispatcher
        Triggered when a file has been added, passing the file object
    file_completed : resumable.util.CallbackDispatcher
        Triggered when a file upload has completed, passing the file object
    chunk_completed : resumable.util.CallbackDispatcher
        Triggered when a chunk upload has completed, passing the file and chunk
        objects
    """

    def __init__(
        self,
        target: str,
        chunk_size: int = MiB,
        simultaneous_uploads: int = 3,
        headers: Optional[Dict[str, Any]] = None,
        test_chunks: bool = True,
        max_chunk_retries: int = 100,
        permanent_errors: Sequence[int] = (400, 404, 415, 500, 501),
        query: Optional[Dict[str, Any]] = None,
        generate_unique_identifier: Callable[
            [Path, Path], str
        ] = lambda _path, _relative_path: str(uuid.uuid4()),
        client: httpx.Client = httpx.Client(),
    ) -> None:
        if headers is None:
            headers = {}
        if query is None:
            query = {}

        self.config = Config(
            target=target,
            chunk_size=chunk_size,
            simultaneous_uploads=simultaneous_uploads,
            headers=headers,
            test_chunks=test_chunks,
            max_chunk_retries=max_chunk_retries,
            permanent_errors=permanent_errors,
            additional_query_params=query,
            generate_unique_identifier=generate_unique_identifier,
        )

        self.client = client
        self.client.headers.update(headers)

        self.files: List[ResumableFile] = []

        self.executor = ThreadPoolExecutor(simultaneous_uploads)
        self.futures: List[Future] = []

        self.file_added = CallbackDispatcher()
        self.file_completed = CallbackDispatcher()
        self.chunk_completed = CallbackDispatcher()

    def add_file(
        self, path: Union[PathLike, str], relative_path: Optional[Path]
    ) -> ResumableFile:
        file = ResumableFile(
            Path(path),
            relative_path,
            self.config.chunk_size,
            self.config.generate_unique_identifier,
        )
        self.files.append(file)

        self.file_added.trigger(file)
        file.completed.register(partial(self.file_completed.trigger, file))
        file.chunk_completed.register(partial(self.chunk_completed.trigger, file))

        for chunk in file.chunks:
            future = self.executor.submit(
                resolve_chunk, self.client, self.config, file, chunk
            )
            self.futures.append(future)

        return file

    def _wait(self) -> None:
        """Wait until all current uploads are completed."""
        for future in as_completed(self.futures):
            exception = future.exception()
            if exception is not None:
                raise exception

    def _cancel_remaining_futures(self) -> None:
        for future in self.futures:
            if not future.done():
                future.cancel()

    def join(self) -> None:
        """Block until all uploads are complete, or an error occurs."""
        try:
            self._wait()
        except:  # noqa: E722
            self._cancel_remaining_futures()
            raise
        finally:
            self.executor.shutdown()
            for file in self.files:
                file.close()

    def __enter__(self) -> "Resumable":
        return self

    def __exit__(self, *args: Any, **kwargs: Any) -> None:
        self.join()



import json
from typing import TYPE_CHECKING, Any, Dict

import numpy as np

from .data_format import DataFormat
from .defaults import ZARR_JSON_FILE_NAME, ZATTRS_FILE_NAME, ZGROUP_FILE_NAME

if TYPE_CHECKING:
    from .dataset import Dataset
    from .layer import Layer


def get_ome_0_4_multiscale_metadata(
    dataset: "Dataset", layer: "Layer"
) -> Dict[str, Any]:
    return {
        "multiscales": [
            {
                "version": "0.4",
                "axes": [
                    {"name": "c", "type": "channel"},
                    {
                        "name": "x",
                        "type": "space",
                        "unit": "nanometer",
                    },
                    {
                        "name": "y",
                        "type": "space",
                        "unit": "nanometer",
                    },
                    {
                        "name": "z",
                        "type": "space",
                        "unit": "nanometer",
                    },
                ],
                "datasets": [
                    {
                        "path": mag.path.name,
                        "coordinateTransformations": [
                            {
                                "type": "scale",
                                "scale": [1.0]
                                + (
                                    np.array(dataset.voxel_size) * mag.mag.to_np()
                                ).tolist(),
                            }
                        ],
                    }
                    for mag in layer.mags.values()
                ],
            }
        ]
    }


def write_ome_0_4_metadata(dataset: "Dataset", layer: "Layer") -> None:
    if layer.data_format == DataFormat.Zarr3:
        with (layer.path / ZARR_JSON_FILE_NAME).open("w", encoding="utf-8") as outfile:
            json.dump(
                {
                    "zarr_format": 3,
                    "node_type": "group",
                    "attributes": get_ome_0_4_multiscale_metadata(dataset, layer),
                },
                outfile,
                indent=4,
            )
    if layer.data_format == DataFormat.Zarr:
        with (layer.path / ZGROUP_FILE_NAME).open("w", encoding="utf-8") as outfile:
            json.dump({"zarr_format": "2"}, outfile, indent=4)
        with (layer.path / ZATTRS_FILE_NAME).open("w", encoding="utf-8") as outfile:
            json.dump(
                get_ome_0_4_multiscale_metadata(dataset, layer),
                outfile,
                indent=4,
            )



from typing import NamedTuple, Optional
from xml.etree.ElementTree import Element

from loxun import XmlWriter

from .utils import as_int_unless_none, enforce_not_null


class Branchpoint(NamedTuple):
    id: int
    time: Optional[int]  # unix timestamp

    def _dump(self, xf: XmlWriter) -> None:
        props = {"id": str(self.id)}
        if self.time is not None:
            props["time"] = str(self.time)
        xf.tag(
            "branchpoint",
            props,
        )

    @classmethod
    def _parse(cls, nml_branchpoint: Element) -> "Branchpoint":
        return cls(
            int(enforce_not_null(nml_branchpoint.get("id"))),
            as_int_unless_none(nml_branchpoint.get("time")),
        )



import logging
from typing import List, Optional, Union

import attr

from ..annotation import Annotation, AnnotationInfo
from ..client.api_client.models import (
    ApiBoundingBox,
    ApiExperience,
    ApiNmlTaskParameters,
    ApiTask,
    ApiTaskCreationResult,
    ApiTaskParameters,
    ApiTaskType,
)
from ..client.context import _get_api_client
from ..dataset.dataset import RemoteDataset
from ..geometry import BoundingBox, Vec3Int
from ..utils import warn_deprecated
from .project import Project

logger = logging.getLogger(__name__)


@attr.frozen
class TaskStatus:
    pending_instance_count: int
    active_instance_count: int
    finished_instance_count: int

    @property
    def open_instance_count(self) -> int:
        warn_deprecated("open_instance_count", "pending_instance_count")
        return self.pending_instance_count


@attr.frozen
class TaskType:
    task_type_id: str
    name: str
    description: str
    team_id: str
    team_name: str

    @classmethod
    def _from_api_task_type(cls, api_task_type: ApiTaskType) -> "TaskType":
        return cls(
            api_task_type.id,
            api_task_type.summary,
            api_task_type.description,
            api_task_type.team_id,
            api_task_type.team_name,
        )


@attr.frozen
class Task:
    """Data class containing information about a WEBKNOSSOS task"""

    task_id: str
    project_id: str
    dataset_name: str
    status: TaskStatus
    task_type: TaskType

    @classmethod
    def get_by_id(cls, task_id: str) -> "Task":
        """Returns the task specified by the passed id (if your token authorizes you to see it)"""
        client = _get_api_client(enforce_auth=True)
        api_task = client.task_info(task_id)
        return cls._from_api_task(api_task)

    @classmethod
    def create_from_annotations(
        cls,
        task_type_id: str,
        project_name: str,
        base_annotations: List[Annotation],
        needed_experience_domain: str,
        needed_experience_value: int,
        instances: int = 1,
        script_id: Optional[str] = None,
        bounding_box: Optional[BoundingBox] = None,
    ) -> List["Task"]:
        """Submits tasks in WEBKNOSSOS based on existing annotations, and returns the Task objects"""

        assert (
            len(base_annotations) > 0
        ), "Must supply at least one base annotation to create tasks"

        client = _get_api_client(enforce_auth=True)
        nml_task_parameters = ApiNmlTaskParameters(
            task_type_id=task_type_id,
            needed_experience=ApiExperience(
                domain=needed_experience_domain,
                value=needed_experience_value,
            ),
            pending_instances=instances,
            project_name=project_name,
            script_id=script_id,
            bounding_box=ApiBoundingBox(
                bounding_box.topleft.to_tuple(),
                bounding_box.size.x,
                bounding_box.size.y,
                bounding_box.size.z,
            )
            if bounding_box is not None
            else None,
        )
        annotation_files = [
            (f"{a.name}.zip", a._binary_zip()) for a in base_annotations
        ]
        result = client.tasks_create_from_files(nml_task_parameters, annotation_files)

        return cls._handle_task_creation_result(result)

    @classmethod
    def create(
        cls,
        task_type_id: str,
        project_name: str,
        dataset_name: Union[str, RemoteDataset],
        needed_experience_domain: str,
        needed_experience_value: int,
        starting_position: Vec3Int,
        starting_rotation: Vec3Int = Vec3Int(0, 0, 0),
        instances: int = 1,
        script_id: Optional[str] = None,
        bounding_box: Optional[BoundingBox] = None,
    ) -> List["Task"]:
        """Submits tasks in WEBKNOSSOS based on a dataset, starting position + rotation, and returns the Task objects"""

        client = _get_api_client(enforce_auth=True)
        if isinstance(dataset_name, RemoteDataset):
            dataset_name = dataset_name._dataset_name
        task_parameters = ApiTaskParameters(
            task_type_id=task_type_id,
            needed_experience=ApiExperience(
                domain=needed_experience_domain,
                value=needed_experience_value,
            ),
            pending_instances=instances,
            project_name=project_name,
            script_id=script_id,
            data_set=dataset_name,
            edit_position=starting_position.to_tuple(),
            edit_rotation=starting_rotation.to_tuple(),
            bounding_box=ApiBoundingBox(
                bounding_box.topleft.to_tuple(),
                bounding_box.size.x,
                bounding_box.size.y,
                bounding_box.size.z,
            )
            if bounding_box is not None
            else None,
        )

        response = client.tasks_create([task_parameters])

        return cls._handle_task_creation_result(response)

    @classmethod
    def _from_api_task(cls, api_task: ApiTask) -> "Task":
        return cls(
            api_task.id,
            api_task.project_id,
            api_task.data_set,
            TaskStatus(
                api_task.status.pending,
                api_task.status.active,
                api_task.status.finished,
            ),
            TaskType._from_api_task_type(api_task.type),
        )

    def get_annotation_infos(self) -> List[AnnotationInfo]:
        """Returns AnnotationInfo objects describing all task instances that have been started by annotators for this task"""
        client = _get_api_client(enforce_auth=True)
        api_annotations = client.annotation_infos_by_task(self.task_id)
        return [AnnotationInfo._from_api_annotation(a) for a in api_annotations]

    def get_project(self) -> Project:
        """Returns the project this task belongs to"""
        return Project.get_by_id(self.project_id)

    @classmethod
    def _handle_task_creation_result(
        cls, result: ApiTaskCreationResult
    ) -> List["Task"]:
        if len(result.warnings) > 0:
            logger.warning(
                f"There were {len(result.warnings)} warnings during task creation:"
            )
        for warning in result.warnings:
            logger.warning(warning)

        successes = []
        errors = []
        for t in result.tasks:
            if t.success is not None:
                successes.append(t.success)
            if t.error is not None:
                errors.append(t.error)
        if len(errors) > 0:
            logger.error(f"{len(errors)} tasks could not be created:")
            for error in errors:
                logger.error(error)
        if len(successes) > 0:
            logger.info(f"{len(successes)} tasks were successfully created.")
        return [cls._from_api_task(t) for t in successes]



"""
# Authentication & Server Context

When interacting with a WEBKNOSSOS server, you might need to
specify your user token to authenticate yourself.
You can copy your token from

[https://webknossos.org/auth/token](https://webknossos.org/auth/token)

Using the same methods, you can also specify the webknossos-server if you
are not using the default [webknossos.org](https://webknossos.org) instance,
as well as a timeout for network requests (default is 30 minutes).

There are the following four options to specify which server context to use:

1. Specifying the context in code using the `webknossos_context`
   contextmanager in a `with` statement:

    ```python
    with webknossos_context(token="my_webknossos_token"):
       # code that interacts with webknossos
    ```

    For more information about the `with` statement and contextmanagers,
    please see [this tutorial](https://realpython.com/python-with-statement).

2. You may specify your settings as environment variables `WK_TOKEN` and `WK_URL`:

    ```shell
    WK_TOKEN="my_webknossos_token" python my-script.py
    ```
3. You can also specify those environment variables in a `.env` file
   in your working directory.
   Environment variables set in the command line take precedence.

    ```shell
    # content of .env
    WK_TOKEN="my_webknossos_token"
    WK_URL="…"
    WK_TIMEOUT="3600"  # in seconds
    ```

4. If nothing else is specified and authentication is needed,
   you are asked interactively for a token, which is used for
   subsequent interactions in the same python run as well.
"""


import os
from contextlib import ContextDecorator
from contextvars import ContextVar, Token
from functools import lru_cache
from typing import Any, List, Optional

import attr
from dotenv import load_dotenv
from rich.prompt import Prompt

from ._defaults import DEFAULT_HTTP_TIMEOUT, DEFAULT_WEBKNOSSOS_URL
from .api_client import DatastoreApiClient, WkApiClient

load_dotenv()


@lru_cache(maxsize=None)
def _cached_ask_for_token(webknossos_url: str) -> str:
    # TODO  pylint: disable=fixme
    # -validate token and ask again if necessary
    # -ask if the token should be saved in some .env file
    # -reset invalid tokens
    #  (e.g. use cachetools for explicit cache managment:
    #  https://cachetools.readthedocs.io/en/stable/#memoizing-decorators)
    return Prompt.ask(
        f"\nPlease enter your webknossos token as shown on {webknossos_url}/auth/token ",
        password=True,
    )


@lru_cache(maxsize=None)
def _cached_get_org(context: "_WebknossosContext") -> str:
    current_api_user = context.api_client_with_auth.user_current()
    return current_api_user.organization


# TODO reset invalid tokens e.g. using cachetools  pylint: disable=fixme
@lru_cache(maxsize=None)
def _cached_get_datastore_token(context: "_WebknossosContext") -> str:
    api_datastore_token = context.api_client_with_auth.token_generate_for_data_store()
    return api_datastore_token.token


@lru_cache(maxsize=None)
def _cached__get_api_client(
    webknossos_url: str,
    token: Optional[str],
    timeout: int,
) -> WkApiClient:
    """Generates a client which might contain an x-auth-token header."""
    if token is None:
        return WkApiClient(base_wk_url=webknossos_url, timeout_seconds=timeout)
    return WkApiClient(
        base_wk_url=webknossos_url,
        headers={"X-Auth-Token": token},
        timeout_seconds=timeout,
    )


def _clear_all_context_caches() -> None:
    _cached_ask_for_token.cache_clear()
    _cached_get_org.cache_clear()
    _cached_get_datastore_token.cache_clear()
    _cached__get_api_client.cache_clear()


@attr.frozen
class _WebknossosContext:
    url: str = os.environ.get("WK_URL", default=DEFAULT_WEBKNOSSOS_URL).rstrip("/")
    token: Optional[str] = os.environ.get("WK_TOKEN", default=None)
    timeout: int = int(os.environ.get("WK_TIMEOUT", default=DEFAULT_HTTP_TIMEOUT))

    # all properties are cached outside to allow re-usability
    # if same context is instantiated twice
    @property
    def required_token(self) -> str:
        if self.token is None:
            token = _cached_ask_for_token(self.url)
            # We replace the current context, but leave all previous ones as-is.
            # Any opened contextmanagers will still close correctly, as the stored
            # tokens still point to the correct predecessors.
            _webknossos_context_var.set(
                _WebknossosContext(self.url, token, self.timeout)
            )
            return token
        else:
            return self.token

    @property
    def organization_id(self) -> str:
        return _cached_get_org(self)

    @property
    def datastore_token(self) -> Optional[str]:
        if self.token is None:
            return None
        else:
            return _cached_get_datastore_token(self)

    @property
    def datastore_required_token(self) -> str:
        return _cached_get_datastore_token(self)

    @property
    def api_client(self) -> WkApiClient:
        return _cached__get_api_client(self.url, self.token, self.timeout)

    @property
    def api_client_with_auth(self) -> WkApiClient:
        return _cached__get_api_client(self.url, self.required_token, self.timeout)

    def get_datastore_api_client(self, datastore_url: str) -> DatastoreApiClient:
        return DatastoreApiClient(
            datastore_base_url=datastore_url, timeout_seconds=self.timeout
        )


_webknossos_context_var: ContextVar[_WebknossosContext] = ContextVar(
    "_webknossos_context_var", default=_WebknossosContext()
)


def _get_context() -> _WebknossosContext:
    return _webknossos_context_var.get()


class webknossos_context(ContextDecorator):
    """"""

    def __init__(
        self,
        url: Optional[str] = None,
        token: Optional[str] = None,
        timeout: Optional[int] = None,
    ) -> None:
        """Returns a new WEBKNOSSOS server contextmanager. Use with the `with` statement:
        ```python
        with webknossos_context(token="my_webknossos_token"):
            # code that interacts with webknossos
        ```

        Or as a decorator:
        ```python
        @webknossos_context(token="my_webknossos_token")
        def my_func():
            # code that interacts with webknossos
        ```

        You can specify the following arguments:
        * `url`, by default [https://webknossos.org](https://www.webknossos.org),
        * `token`, as displayed on [https://webknossos.org/auth/token](https://webknossos.org/auth/token),
        * `timeout` to specify a custom network request timeout in seconds, `1800` (30min) by default.

        `url` and `timeout` are taken from the previous context (e.g. environment variables) if not specified.
        `token` must be set explicitly, it is not available when not specified.
        """
        self._url = _get_context().url if url is None else url.rstrip("/")
        self._token = token
        self._timeout = _get_context().timeout if timeout is None else timeout
        self._context_var_token_stack: List[Token[_WebknossosContext]] = []

    def __enter__(self) -> None:
        context_var_token = _webknossos_context_var.set(
            _WebknossosContext(self._url, self._token, self._timeout)
        )
        self._context_var_token_stack.append(context_var_token)

    def __exit__(self, *exc: Any) -> None:
        _webknossos_context_var.reset(self._context_var_token_stack.pop())


def _get_api_client(enforce_auth: bool = False) -> WkApiClient:
    if enforce_auth:
        return _get_context().api_client_with_auth
    else:
        return _get_context().api_client
